import os
import warnings
from dataclasses import dataclass, astuple
from datetime import datetime, timezone as dt_timezone
from typing import Optional

from packaging.version import parse as parse_version

import onetick.py as otp
from onetick.py.otq import otq, otli, pyomd
from onetick.py.backports import cache


@dataclass
class OnetickVersion:
    is_release: bool
    release_version: Optional[str]
    update_number: Optional[int]
    build_number: int


@dataclass
class OnetickVersionFromServer(OnetickVersion):
    db: str
    context: str


def _parse_update_info(update_info: str) -> Optional[int]:
    if update_info == 'initial':
        return 0
    if update_info == 'precandidate':
        return None
    prefix = 'update'
    if not update_info.startswith(prefix):
        raise ValueError(f"Unexpected update info format: '{update_info}'")
    update_info = update_info[len(prefix):]
    return int(update_info)


def _compare_build_string_and_number(build_string: str, build_number: int,
                                     release_format_version: int, release_string: str):
    if release_format_version == 2:
        build_string += '120000'
    try:
        release_build_number = int(build_string)
    except Exception:
        raise ValueError(f"Unexpected build number '{build_string}' in release string '{release_string}'")

    if str(release_build_number) != str(build_number):
        raise ValueError(
            f"Different build numbers in OneTick release '{release_string}' and version: '{build_number}'"
        )


def _parse_release_string(release_string: str, build_number: int) -> OnetickVersion:
    # pylint: disable=W0707

    # Known release string formats:
    #  dev_build
    #  rel_1_23_20230605193357
    #  BUILD_initial_20230831120000
    #  BUILD_update1_20230831120000
    #  BUILD_pre_candidate_20240501000000
    #
    #  BUILD_rel_20241018_initial
    #  BUILD_rel_20241018_update3
    #  rel_1_25_initial
    #  rel_1_25_update1

    if release_string == 'dev_build':
        return OnetickVersion(False, None, None, build_number)

    release_type, *release_info, release_suffix = release_string.split('_')

    if not release_info:
        raise ValueError("No release info")

    try:
        update_number = _parse_update_info(release_suffix)
        release_format_version = 2
    except ValueError:
        update_number = None
        release_format_version = 1
        _compare_build_string_and_number(release_suffix, build_number, release_format_version, release_string)

    if release_type == 'rel':
        release_version_string = '.'.join(release_info)
        release_version = parse_version(release_version_string)
        return OnetickVersion(True, str(release_version), update_number, build_number)

    if release_type == 'BUILD':
        if release_format_version == 1:
            update_info = ''.join(release_info)
            update_number = _parse_update_info(update_info)
        if release_format_version == 2:
            assert release_info[0] == 'rel', 'Unknown release type'
            release_info = release_info[1:]
            build_string = ''.join(release_info)
            _compare_build_string_and_number(build_string, build_number, release_format_version, release_string)

        return OnetickVersion(False, None, update_number, build_number)

    raise ValueError(f"Unknown release type '{release_type}' in release string '{release_string}'")


def _get_locator_intervals(db_name, context) -> list[tuple[datetime, datetime]]:
    graph = otq.GraphQuery(otq.DbShowConfiguredTimeRanges(db_name=db_name).tick_type('ANY')
                           >> otq.Table(fields='long START_DATE, long END_DATE'))
    symbols = f'{db_name}::'

    # setting this is important so we don't get access error
    qp = pyomd.QueryProperties()
    qp.set_property_value('IGNORE_TICKS_IN_UNENTITLED_TIME_RANGE', 'TRUE')

    result = otq.run(graph,
                     symbols=symbols,
                     # start and end times don't matter for this query, use some constants
                     start=datetime(2003, 12, 1),
                     end=datetime(2003, 12, 1),
                     # timezone is irrelevant, because times are returned as epoch numbers
                     timezone='UTC',
                     query_properties=qp,
                     context=context)
    data = result.output(symbols).data
    if not data:
        raise RuntimeError(f"Database '{db_name}' doesn't have locations")
    return [
        (
            datetime.fromtimestamp(data['START_DATE'][i] / 1000, dt_timezone.utc).replace(tzinfo=None),
            datetime.fromtimestamp(data['END_DATE'][i] / 1000, dt_timezone.utc).replace(tzinfo=None),
        )
        for i in range(len(data['START_DATE']))
    ]


def _get_onetick_version(db_name, context, start, end) -> dict:
    node = otq.TickGenerator(bucket_interval=0,
                             fields='BUILD=GET_ONETICK_VERSION(), RELEASE=GET_ONETICK_RELEASE()')
    graph = otq.GraphQuery(node.tick_type('DUMMY'))
    symbols = f'{db_name}::'

    # setting this is important so we don't get access error
    qp = pyomd.QueryProperties()
    qp.set_property_value('IGNORE_TICKS_IN_UNENTITLED_TIME_RANGE', 'TRUE')

    result = otq.run(graph,
                     symbols=symbols,
                     start=start,
                     end=end,
                     context=context,
                     query_properties=qp,
                     timezone='UTC')
    data = result.output(symbols).data
    if not data:
        raise RuntimeError(f"Can't get OneTick version from database '{db_name}'")
    return data


@cache
def get_onetick_version(db=None, context=None) -> OnetickVersionFromServer:
    """
    Get OneTick release version, as build number isn't enough
    to determine features available in OneTick.

    Returns tuple with release type, release version, update number and build number.

    Note
    ----
    The version is taken from the server by calling the query against this server.

    The server is specified by two global configuration parameters:
    :py:attr:`otp.config.context<onetick.py.configuration.Config.context>`
    and :py:attr:`otp.config.default_db<onetick.py.configuration.Config.default_db>`.
    By default, 'DEFAULT' context and 'LOCAL' database will be used.

    The check will not be accurate in all cases, as the user may use :func:`otp.run <onetick.py.run>`
    with different context or set symbol with different database in the end.

    Checking version correctly in all cases requires redesigning compatibility check system
    by moving it to the runtime level -- checking version inside the graph.
    But for now this method is the best we can do.
    """
    s = None
    if not os.environ.get('ONE_TICK_CONFIG') and not otq.webapi:
        s = otp.Session()
    else:
        _ = otli.OneTickLib()

    # if otp.config.default_db is set, then we use it to check compatibility
    # otherwise we use LOCAL database available everywhere
    db_name = db or otp.config.get('default_db', 'LOCAL')
    context = context or otp.config.context

    try:
        if db_name == 'LOCAL':
            # for LOCAL db any date will do
            start = end = datetime(2003, 12, 1)
            result_data = _get_onetick_version(db_name, context, start, end)
        else:
            # for real db we need to set time range correctly
            # otherwise we may get error "Database locator has a gap"
            locator_intervals = _get_locator_intervals(db_name, context)
            for i, (start, end) in enumerate(locator_intervals):
                try:
                    result_data = _get_onetick_version(db_name, context, start, end)
                    break
                except Exception as e:
                    if i < len(locator_intervals) - 1:
                        continue
                    else:
                        raise e
    finally:
        if s:
            s.close()

    build_number = int(result_data["BUILD"][0])
    release_string = result_data["RELEASE"][0]

    try:
        onetick_version = _parse_release_string(release_string, build_number=build_number)
        return OnetickVersionFromServer(*astuple(onetick_version), db, context)  # type: ignore[call-arg]
    except Exception as err:
        warnings.warn(f"Unknown release format string: '{release_string}'.\n{err}")
        return OnetickVersionFromServer(False, None, None, build_number, db, context)


def _is_min_build_or_version(min_release_version=None,
                             min_release_version_build_number=None,
                             min_build_number=None,
                             min_update_number=None,
                             throw_warning=False,
                             feature_name=None,
                             db=None,
                             context=None):
    """
    Check if current OneTick version is at least min_release_version.
    When using not released version, check if build number is at least min_build_number.
    """
    if not min_build_number:
        raise ValueError("min_build_number parameter is required")

    from onetick.py.configuration import config
    if config.disable_compatibility_checks:
        return True

    onetick_version = get_onetick_version(db=db, context=context)
    if not onetick_version.is_release:
        has = onetick_version.build_number >= min_build_number
        if (
            min_update_number is not None
            and onetick_version.update_number is not None
            and onetick_version.build_number == min_build_number
        ):
            has = has and onetick_version.update_number >= min_update_number
    else:
        if not min_release_version:
            # onetick is on release, but feature is not released yet
            has = False
        else:
            has = parse_version(str(onetick_version.release_version)) >= parse_version(str(min_release_version))
            if min_release_version_build_number:
                has = has and onetick_version.build_number >= min_release_version_build_number

    if not has and throw_warning:
        msg = f"OneTick {onetick_version} does not support {feature_name} which is supported "
        if min_release_version is not None:
            msg += f"starting from release {min_release_version} or "
        msg += f"starting from dev build {min_build_number} "
        if min_update_number is not None:
            msg += f"update {min_update_number}"
        warnings.warn(msg)
    return has


def _add_version_info_to_exception(exc):
    """
    Add onetick-py and onetick version numbers to exception message.
    """
    onetick_version = get_onetick_version()
    if not onetick_version.is_release:
        message = f'OneTick {onetick_version.build_number}'
    else:
        message = f'OneTick {onetick_version.release_version} ({onetick_version.build_number})'
    message = f'onetick-py=={otp.__version__}, {message}'
    if exc.args:
        message = str(exc.args[0]) + os.linesep + message
    exc.args = (message, *exc.args[1:])
    return exc


def has_max_expected_ticks_per_symbol(throw_warning=False):
    """Check if otq.run() has max_expected_ticks_per_symbol parameter.

    20220531: Implemented 0027950: OneTick numpy API and onetick.query python API
    should expose parameter max_expected_ticks_per_symbol
    """
    has = _is_min_build_or_version(1.23, 20221025023710,
                                   20220714120000,
                                   throw_warning=throw_warning,
                                   feature_name="otp.run parameter 'max_expected_ticks_per_symbol'")
    return has


def has_password_param(throw_warning=False):
    """Check if otq.run() has password parameter.

    Implemented 0027216: onetick.query does not expose parameter password
    """
    has = _is_min_build_or_version(1.23, None,
                                   20220327120000,
                                   throw_warning=throw_warning,
                                   feature_name="otp.run parameter 'password'")
    return has


def has_timezone_parameter(throw_warning=False):
    """
    Fixed 0027499: In onetick.query, _convert_time_to_YYYYMMDDhhmmss method should accept timezone parameter
    """
    has = _is_min_build_or_version(1.23, None,
                                   20220519120000,
                                   throw_warning=throw_warning,
                                   feature_name="convert_time_to_YYYYMMDDhhmmss parameter 'timezone'")
    return has


def has_query_encoding_parameter(throw_warning=False):
    """
    0027383: In onetick.query, run method should support parameter "encoding"
    """
    has = _is_min_build_or_version(1.23, None,
                                   20220327120000,
                                   throw_warning=throw_warning,
                                   feature_name="query encoding parameter")
    return has


def is_supported_agg_option_price():
    """0029945: OPTION_PRICE EP produces an exception when used in COMPUTE and explicitly set to its default value
    """
    return _is_min_build_or_version(1.23, 20230314061408,
                                    20230316120000)


def is_supported_otq_run_password():
    """Implemented 0027216: onetick.query does not expose parameter password
    """
    return _is_min_build_or_version(1.23, None,
                                    20220327120000)


def is_supported_stack_info():
    """Fixed 0028824: setting otq.API_CONFIG.SHOW_STACK_INFO=1 does not cause location of an EP in
    python code to be added to the text of exception
    """
    onetick_version = get_onetick_version()
    if onetick_version.build_number == 20240205120000:
        # BDS-345
        return False
    return _is_min_build_or_version(1.24, None,
                                    20221111120000)


def is_supported_num_distinct():
    """???
    """
    return _is_min_build_or_version(1.23, None,
                                    20220913120000)


def is_supported_rename_fields_symbol_change():
    """???
    """
    return _is_min_build_or_version(1.24, 20230316120000,
                                    20230316120000)


def is_supported_new_ob_snapshot_behavior():
    """???
    """
    return _is_min_build_or_version(1.24, 20230711120000,
                                    20230711120000)


def is_supported_where_clause_for_back_ticks():
    """Implemented 0028064: add WHERE_CLAUSE_FOR_BACK_TICKS to PASSTHROUGH EP
    """
    return _is_min_build_or_version(1.23, None,
                                    20220714120000)


def is_supported_bucket_units_for_tick_generator(throw_warning=False):
    """Implemented 0029117: Add BUCKET_INTERVAL_UNITS to TICK_GENERATOR EP
    """
    feature_name = "parameter 'bucket_units' for otp.Tick"
    return _is_min_build_or_version(1.24, None,
                                    20230112120000,
                                    throw_warning=throw_warning, feature_name=feature_name)


def is_supported_varstring_in_get_string_value():
    """Implemented 0030763: GET_STRING_VALUE method on tick objects should support also varstring field types
    """
    return _is_min_build_or_version(1.24, None,
                                    20230711120000)


def is_supported_uint_numpy_interface():
    # 20220216: Fixed 0027130: onetick numpy interface should preserve
    # field type of unsigned fields (currently they become signed)
    return _is_min_build_or_version(1.23, None,
                                    20220327120000)


def is_supported_otq_reference_data_loader():
    """???
    """
    return _is_min_build_or_version(1.23, 20220519120000, 20220519120000)


def is_supported_nsectime_tick_set_eval():
    # BDS-321
    # Fixed 0031588: Ticks in TICK_SET populated by eval , loose nanosecond precision
    return _is_min_build_or_version(1.24, None,
                                    20231108120000)


def is_supported_otq_ob_summary(throw_warning=False):
    """
    20220325: Implemented 0027258: Add EP OB_SUMMARY, which will combine functionality of OB_SIZE, OB_NUM_LEVELS, and
    OB_VWAP, and add new features
    """
    has = _is_min_build_or_version(1.23, None,
                                   20220327120000,
                                   throw_warning=throw_warning,
                                   feature_name="onetick.query OB_SUMMARY support")
    return has


def is_supported_reload_locator_with_derived_db():
    # See tasks PY-388, BDS-334.
    # Was fixed in update1_20231108120000.
    # 0032118: OneTick processes that refresh their locator may crash
    #          if they make use databases derived from the dbs in that locator
    return _is_min_build_or_version(1.24, None,
                                    20231108120000, min_update_number=1)


def is_supported_large_ints_empty_interval():
    # BDS-333
    # Was fixed in update1_20231108120000.
    # 0032093: when EXPECT_LARGE_INTS isn't 'false, HIGH,LOW,FIRST, and LAST EPs should show integer values,
    #          not doubles, when input is empty
    return _is_min_build_or_version(1.24, None,
                                    20231108120000, min_update_number=1)


def is_start_time_as_minimum_start_date_supported():
    # 20220203: Fixed 0027114: Getting error when query start time is equal
    # to <minimum_start_date> parameter in access control file.
    return _is_min_build_or_version(1.23, None,
                                    20220211120000)


def is_supported_list_empty_derived_databases():
    # PY-856, BDS-323
    # Was fixed in BUILD_initial_20240205120000
    # 20240130: Fixed 0031783: onetick.query crashes when a query returned no ticks,
    # but produced a tick descriptor with string fields of 0 size
    return _is_min_build_or_version(1.24, 20240524004422,
                                    20240205120000, min_update_number=0)


def is_odbc_query_supported():
    # no record found in Release Notes
    # but grep shows that it was added in 20231108-0 build and 1.24 release
    return _is_min_build_or_version(1.24, None,
                                    20231108120000)


def is_event_processor_repr_upper():
    if otq.webapi:
        return True
    return _is_min_build_or_version(1.25, None,
                                    20240205120000, min_update_number=0)


def is_date_trunc_fixed():
    # Fixed 0032253: DATE_TRUNC function returns wrong answer in case of daylight saving time
    return _is_min_build_or_version(1.25, None,
                                    20240205120000, min_update_number=0)


def is_supported_end_time_in_modify_state_var_from_query():
    # BDS-335 [onetick 0032075]: End time for the called query in MODIFY_STATE_VAR_FROM_QUERY is set incorrectly
    # Was fixed in update1_20231108120000.
    return _is_min_build_or_version(1.24, None,
                                    20231108120000, min_update_number=1)


def is_supported_modify_state_var_from_query():
    return hasattr(otq, 'ModifyStateVarFromQuery')


def is_sha2_hashing_supported():
    return _is_min_build_or_version(1.23, 20220714120000,
                                    20220714120000)


def is_supported_join_with_aggregated_window():
    return hasattr(otq, 'JoinWithAggregatedWindow')


def is_existing_fields_handling_supported():
    # 20220207: Implemented 0027076:
    # ADD_FIELDS should support parameter EXISTING_FIELDS_HANDLING with values THROW and OVERRIDE
    return _is_min_build_or_version(1.23, None,
                                    20220211120000)


def is_supported_per_cache_otq_params(throw_warning=False):
    return _is_min_build_or_version(1.23, 20220714120000, 20220714120000, throw_warning=throw_warning)


def is_option_price_theta_value_changed():
    # 20240221: Fixed 0032506:
    # Theta value from OPTION_PRICE EP is sometimes wrong.
    return _is_min_build_or_version(1.24, 20240306230425,
                                    20240330120000)


def is_fixed_modify_state_var_from_query():
    # 20230913: Fixed 0031340:
    # MODIFY_STATE_VAR_FROM_QUERY does not properly propagate initialization events
    # which may cause crash in destination EPs
    return _is_min_build_or_version(1.24, None,
                                    20231108120000)


def is_supported_next_in_join_with_aggregated_window(throw_warning=False, feature_name=None):
    # 20231111: Fixed 0031756:
    # Queries with JOIN_WITH_AGGREGATED_WINDOW crash
    # if it is followed by Aggregation EPs referencing fields in PASS_SOURCE
    return _is_min_build_or_version(1.24, None,
                                    20231108120000, min_update_number=1,
                                    throw_warning=throw_warning, feature_name=feature_name)


def is_min_db_start_criteria_works_correctly():
    # Works from 1.23
    # 20220512: Implemented 0027673:
    # SHOW_SYMBOLOGY_LIST should not throw start/end date criteria violation exceptions
    return _is_min_build_or_version(1.23, 20220519120000,
                                    20220519120000)


def is_repeat_with_field_name_works_correctly():
    # Works before 20230522-0, on 20230522-2/4 and after 20230711
    # 20230705: Fixed 0030642:
    # built-in REPEAT function works incorrectly when passed a field name
    # as opposed to the constant string, starting rel_20230522
    onetick_version = get_onetick_version()

    if (
        onetick_version.build_number < 20230522120000 or
        onetick_version.build_number == 20230522120000 and onetick_version.update_number >= 2 or
        onetick_version.build_number >= 20230711120000
    ):
        return True

    return False


def is_duplicating_quotes_not_supported():
    # 20240329: Fixed 0032754:
    # Logical expressions should trigger error when duplicate single(or double) quote
    # is directly followed or preceded by some name
    return _is_min_build_or_version(1.25, None,
                                    20240330120000)


def are_quotes_in_query_params_supported():
    # Fixed 0033318: onetick.query package passes quoted otq parameters without quotes
    return _is_min_build_or_version(None, None, 20240530120000, min_update_number=1)


def is_concurrent_cache_is_fixed():
    # PY-1009, BDS-365
    # 20240802: Fixed 0033806: Dynamic caches created with PER_CACHE_OTQ_PARAMS in READ_CACHE EP
    # still lack synchronization in multi-core environment.
    return _is_min_build_or_version(1.24, 20240806024006,
                                    20240812120000)


def is_apply_rights_supported(throw_warning=False):
    # 20191026: Fixed 0021898: CORP_ACTIONS EP does not expose parameter APPLY_RIGHTS
    return _is_min_build_or_version(1.22, 20220128183755,
                                    20220714120000,
                                    throw_warning=throw_warning)


def is_write_parquet_directories_fixed():
    # 20240609: Fixed 0033342: WRITE_TO_PARQUET EP should not produce directories in non-partitioned mode
    return _is_min_build_or_version(1.25, 20250209162722,
                                    20240530120000, min_update_number=1)


def is_zero_concurrency_supported():
    # 20240312: Implemented 0032157:
    # Add support for automatic assignment of concurrency to the queries, if concurrency is set to special value '0'
    return _is_min_build_or_version(None, None,
                                    20240501000000)


def is_get_query_property_flag_supported():
    # 20231205: Implemented 0031857:
    # create flag for GET_QUERY_PROPERTY and GET_QUERY_PROPERTIES to return also special query properties
    return _is_min_build_or_version(None, None,
                                    20240205120000)


def is_all_fields_when_ticks_exit_window_supported():
    # 20231230: Implemented 0031741:
    # ALL_FIELDS_FOR_SLIDING aggregation parameter should support value WHEN_TICKS_EXIT_WINDOW
    # (check out "Parameters common go generic aggregations" section in OneTick Event Processors' guide).
    return _is_min_build_or_version(1.24, 20240116201311,
                                    20240205120000)


def is_first_ep_skip_tick_if_supported():
    # 20240130: Implemented 0032167: Add SKIP_TICK_IF parameter for FIRST EP
    return _is_min_build_or_version(None, None,
                                    20240205120000)


def is_last_ep_fwd_fill_if_supported():
    # 20220708: Implemented 0028111: LAST EP should have parameter FWD_FILL_IF
    return _is_min_build_or_version(1.23, 20221025023710,
                                    20220714120000)


def is_diff_show_matching_ticks_supported():
    return _is_min_build_or_version(None, None,
                                    20240812120000)


def is_diff_non_decreasing_value_fields_supported():
    # 20240620: Implemmented 0033285: extend DIFF EP to support matching ticks with non-identical primary timestamps
    return _is_min_build_or_version(None, None,
                                    20240812120000)


def is_standardized_moment_supported():
    # 20240513: Implemented 0032822: Add STANDARDIZED_MOMENT EP, to compute STANDARDIZED_MOMENT of Nth degree
    return _is_min_build_or_version(None, None,
                                    20240530120000)


def is_supported_pnl_realized():
    # No info, however onetick.query missing required EP class
    return _is_min_build_or_version(1.24, 20240116201311,
                                    20231108120000)


def is_supported_pnl_realized_buy_sell_flag_bin():
    # 20240429: Implemented 0032683: Enhance PNL_REALIZED EP for BUY_SELL_FLAG field to support also 0 and 1
    return _is_min_build_or_version(None, None,
                                    20240530120000)


def is_data_file_query_supported():
    # 20240311: Implemented 0032631: Implement ARROW_FILE_QUERY EP
    return _is_min_build_or_version(None, None,
                                    20240330120000)


def is_data_file_query_symbology_supported(throw_warning=False, feature_name=None):
    # 20240603: Implemented 0033111: DATA_FILE_QUERY EP should support parameter SYMBOLOGY
    return _is_min_build_or_version(None, None,
                                    20240812120000,
                                    throw_warning=throw_warning, feature_name=feature_name)


def is_supported_point_in_time(throw_warning=False, feature_name=None):
    # 20240323: Implemented 0032255: Add POINT_IN_TIME EP
    # 20240408: Implemented 0032821: enhance POINT_IN_TIME EP to support getting points in time
    # from the input time series, when TIMES parameter is not set.

    # POINT_IN_TIME EP supported since 20240330120000, but it is not very stable in this first version,
    # so we decided to support it since the next version
    return _is_min_build_or_version(1.25, 20241209135932,
                                    20240530120000,
                                    throw_warning=throw_warning, feature_name=feature_name)


def is_find_value_for_percentile_supported():
    # 20240527: Implemented 0032752: Add EP FIND_VALUE_FOR_PERCENTILE
    return _is_min_build_or_version(None, None,
                                    20240530120000)


def is_derived_databases_crash_fixed():
    # 20240130: Fixed 0032118: OneTick processes that refresh their locator
    # may crash if they make use of databases derived from the dbs in that locator
    return _is_min_build_or_version(1.24, 20240524004422,
                                    20240205120000)


def is_character_present_characters_field_fixed():
    # 20230705: Fixed 0030747: CHARACTER_PRESENT EP may produce non-deterministic results when
    # CHARACTERS_FIELD is specified
    # 20230705: Fixed 0030748: CHARACTER_PRESENT EP must ignore 0-bytes in the values of a tick field named
    # by the CHARACTERS_FIELD parameter
    return _is_min_build_or_version(1.24, 20240116201311,
                                    20230711120000)


def is_supported_estimate_ts_delay():
    # 20240924: Implemented 0033286: Add EP ESTIMATE_TS_DELAY
    return _is_min_build_or_version(None, None,
                                    20241002120000)


def is_percentile_bug_fixed():
    # 20241209: Implemented 0034428: In FIND_VALUE_FOR_PERCENTILE EP, rename SHOW_PERCENTILE_AS to COMPUTE_VALUE_AS
    # NOTE: also has fix for FIRST_VALUE_WITH_GE_PERCENTILE and PERCENTILE=100 (was N/A, but must be biggest value)
    return _is_min_build_or_version(None, None,
                                    20241220120000)


def is_limit_ep_supported():
    # Implemented 0034293: LIMIT ep
    return (
        hasattr(otq, 'Limit') and
        _is_min_build_or_version(1.25, 20241229055942,
                                 20241018120000, min_update_number=1)
    )


def is_limit_tick_offset_supported():
    # Implemented OTDEV-37257: LIMIT EP should support TICK_OFFSET parameter
    return (
        is_limit_ep_supported() and
        'tick_offset' in otq.Limit.Parameters.list_parameters() and
        _is_min_build_or_version(None, None,
                                 20251010120000, min_update_number=2)
    )


def is_prefer_speed_over_accuracy_supported(**kwargs):
    return _is_min_build_or_version(1.25, 20241229055942,
                                    20241018120000, min_update_number=3,
                                    **kwargs)


def is_ob_virtual_prl_and_show_full_detail_supported():
    # 20230705: Implemented 0030536: VIRTUAL_OB EP should support PRL output format and should require it
    # for SHOW_FULL_DETAIL case
    return _is_min_build_or_version(1.24, 20240116201311,
                                    20230711120000)


def is_per_tick_script_boolean_problem():
    # strange problem, couldn't reproduce it anywhere except a single onetick release
    version = get_onetick_version()
    return version.release_version == '1.22'


def is_symbol_time_override_fixed():
    # Fixed 0028044: after rel_20220519, symbol_date=0 in the otq file overrides
    # symbol_date expressions and _SYMBOL_TIME otq parameter
    return _is_min_build_or_version(1.23, 20221025023710,
                                    20220714120000)


def is_database_view_schema_supported():
    # Implemented 0034115: DB/SHOW_TICK_TYPES should return non-empty schema
    # for View queries ending in single TABLE EP with type specified for each field
    return _is_min_build_or_version(1.25, 20241229055942,
                                    20241001205534)


def is_native_plus_zstd_supported():
    # 20220204: Implemented 0026827: memory and accelerator databases should support ZSTD, NATIVE_PLUS_ZSTD,
    # and per-tick ZSTD compression
    return _is_min_build_or_version(1.23, 20220913120000,
                                    20220211120000)


def is_save_snapshot_database_parameter_supported():
    # 20220929: Implemented 0028559: Update SAVE_SNAPSHOT to specify output database
    client_support = 'database' in otq.SaveSnapshot.Parameters.list_parameters()
    server_support = _is_min_build_or_version(1.23, 20230605193357,
                                              20221111120000)
    return client_support and server_support


def is_join_with_snapshot_snapshot_fields_parameter_supported():
    # 20240422: Implemented 0032910: add parameter SNAPSHOT_FIELDS to JOIN_WITH_SNAPSHOT EP
    return _is_min_build_or_version(1.25, 20241229055942,
                                    20240530120000)


def is_multi_column_generic_aggregations_supported():
    # Implementation of tick aggregations in COMPUTE requires to use RENAME_FIELDS to make correct output schema.
    # However, if we place it inside generic aggregation inside COMPUTE, next error occur on old OneTick versions:
    # ERR_06708004ERCOM: Event processor RENAME_FIELDS does not currently support dynamic symbol changes.
    return _is_min_build_or_version(1.24, 20240116201311,
                                    20230315095103)


def is_max_concurrency_with_webapi_supported():
    # 0036758: in onetick.query_webapi: max_concurrency is not being saved in otq file when set on otq.Query
    # 0036759: in onetick.query_webapi:
    # it's not possible to pass max_concurrency 0 in method otq.run when using otq file
    return _is_min_build_or_version(None, None,
                                    20250727120000, min_update_number=3)


def is_nanoseconds_fixed_in_run():
    # 0032309: onetick.query_webapi should preserve nanosecond timestamps
    return _is_min_build_or_version(1.24, 20240116201311,
                                    20240205120000)


def is_correct_timezone_used_in_otq_run():
    # Fixed 0027500: In onetick.query, OtqFile.save_to_file method uses incorrect timezone
    return _is_min_build_or_version(1.23, 20221025023710,
                                    20220519120000)


def is_ilike_supported():
    # 20250423: Implemented 0035414: Add support of ILIKE in PER_TICK_SCRIPT EP
    # 20250423: Implemented 0035412: Add support of ILIKE in logical expressions
    # 20250423: Implemented 0035413: Add support of ILIKE in SQL
    return _is_min_build_or_version(None, None,
                                    20250510120000)


def is_include_market_order_ticks_supported(**kwargs):
    # Implemented 0031478: OB_SNAPSHOT... and OB_SUMMARY EPs
    # should support parameter INCLUDE_MARKET_ORDER_TICKS (false by default)
    return _is_min_build_or_version(1.25, 20241229055942,
                                    20240812120000, min_update_number=2,
                                    **kwargs)


def is_join_with_query_symbol_time_otq_supported():
    # 20241209: Fixed 0034770: hours/minutes/seconds part of otq parameter _SYMBOL_TIME, expressed in
    # milliseconds since 1970/01/01 00:00:00 GMT, is ignored
    # 20250219: Implemented 0035092: passing otq param _SYMBOL_TIME should be just like setting symbol_date
    # to the equivalent value, except in YYYYMMDDhhmmss format
    return _is_min_build_or_version(None, None,
                                    20250227120000)


def is_show_db_list_show_description_supported():
    # 20240301: Implemented 0032320: 0032320: SHOW_DB_LIST should have a new EP parameter, SHOW_DESCRIPTION
    # However on 20240330 builds it returns SHOW_DESCRIPTION column instead of DESCRIPTION
    return _is_min_build_or_version(1.25, 20241229055942,
                                    20240501000000)


def is_symbols_prepend_db_name_supported():
    # 20250924: Implemented 0036753: FIND_DB_SYMBOLS should have EP parameter PREPEND_DB_NAME (true by default)
    return hasattr(otq.FindDbSymbols.Parameters, 'prepend_db_name') and _is_min_build_or_version(
        None, None, 20251010120000,
    )


def is_diff_show_all_ticks_supported():
    # 20250919: Implemented 0036784: Add SHOW_ALL_TICKS(false by default) ep parameter to DIFF EP.
    return hasattr(otq.Diff.Parameters, 'show_all_ticks') and _is_min_build_or_version(
        None, None, 20251010120000,
    )


def is_max_spread_supported():
    # 20250819: Implemented 0036522: The book EPs that support parameter MAX_DEPTH_FOR_PRICE
    # should also support parameter MAX_SPREAD
    return _is_min_build_or_version(None, None,
                                    20251010120000)


def is_not_fixed_bds_484():
    # BDS-484: seems like timezone is ignored in otq.run in some cases
    return _is_min_build_or_version(None, None,
                                    20251010120000, min_update_number=2)


def is_webapi_access_token_scope_supported():
    # 20251030: Fixed OTDEV-37063: onetick.query_webapi.get_access_token method must take scope as a parameter
    return _is_min_build_or_version(None, None,
                                    20251010120000, min_update_number=2)


def is_expect_decimals_supported(agg_name: str):
    # 20251218 build
    # 20251203: Implemented OTDEV-37054: Add parameter EXPECT_DECIMALS to LOW(_TICK), HIGH(_TICK), FIRST, and LAST EPs
    # 20251117: Implemented OTDEV-37055: Add parameter EXPECT_DECIMALS to SUM and MEDIAN EPs

    # 1.26 20251218 and 20251010-3 builds
    # 20251208: Implemented OTDEV-37054: Add parameter EXPECT_DECIMALS to LOW(_TICK), HIGH(_TICK), FIRST, and LAST EPs
    if agg_name in ['HIGH', 'LOW', 'FIRST', 'LAST']:
        return _is_min_build_or_version(1.26, 20251217195357,
                                        20251010120000, min_update_number=3)
    elif agg_name in ['SUM', 'MEDIAN']:
        return _is_min_build_or_version(None, None,
                                        20251218120000)

    raise ValueError(f'Unsupported aggregation `{agg_name}`')



from typing import Union

import onetick.py as otp
from onetick.py.core.column import _Column
from .compute import Compute
from .high_low import Max, Min, HighTick, LowTick, HighTime, LowTime
from .other import (First, Last, FirstTime, LastTime, Count, Vwap, FirstTick,
                    LastTick, Distinct, Sum, Average, StdDev, TimeWeightedAvg,
                    Median, Correlation, OptionPrice, Ranking, Variance,
                    Percentile, FindValueForPercentile, ExpWAverage, ExpTwAverage,
                    StandardizedMoment, PortfolioPrice, MultiPortfolioPrice, Return, ImpliedVol,
                    LinearRegression, PartitionEvenlyIntoGroups)
from .order_book import (ObSnapshot, OB_SNAPSHOT_DOC_PARAMS,
                         ObSnapshotWide, OB_SNAPSHOT_WIDE_DOC_PARAMS,
                         ObSnapshotFlat, OB_SNAPSHOT_FLAT_DOC_PARAMS,
                         ObSummary, OB_SUMMARY_DOC_PARAMS,
                         ObSize, OB_SIZE_DOC_PARAMS,
                         ObVwap, OB_VWAP_DOC_PARAMS,
                         ObNumLevels, OB_NUM_LEVELS_DOC_PARAMS)
from .generic import Generic
from ._docs import (_column_doc,
                    _running_doc,
                    _all_fields_doc,
                    _bucket_interval_doc,
                    _bucket_time_doc,
                    _bucket_units_doc,
                    _bucket_end_condition_doc,
                    _end_condition_per_group_doc,
                    _boundary_tick_bucket_doc,
                    _group_by_doc,
                    _groups_to_display_doc,
                    _n_doc,
                    _keep_timestamp_doc,
                    _time_series_type_doc,
                    _time_series_type_w_doc,
                    _selection_doc,
                    _query_fun_doc,
                    _bucket_delimiter_doc,
                    _biased_doc,
                    _large_ints_doc,
                    _null_int_val_doc,
                    _expect_decimals_doc,
                    _skip_tick_if_doc,
                    _default_tick_doc,
                    _decay_doc,
                    _decay_value_type_doc,
                    _decay_value_type_hl_doc,
                    _degree_doc,
                    _weight_field_name_doc,
                    _weight_multiplier_field_name_doc,
                    _portfolio_side_doc,
                    _weight_type_doc,
                    _symbols_doc,
                    _portfolios_query_doc,
                    _portfolios_query_params_doc,
                    _portfolio_value_field_name_doc,
                    _columns_portfolio_doc,
                    _interest_rate_doc,
                    _price_field_doc,
                    _option_price_field_doc,
                    _method_doc,
                    _precision_doc,
                    _value_for_non_converge_doc,
                    _option_type_field_doc,
                    _strike_price_field_doc,
                    _days_in_year_doc,
                    _days_till_expiration_field_doc,
                    _expiration_date_field_doc,
                    _dependent_variable_field_name_doc,
                    _independent_variable_field_name_doc,
                    _field_to_partition_doc,
                    _weight_field_doc,
                    _number_of_groups_doc)

from onetick.py.docs.utils import docstring, param_doc

# This module mostly focused on providing annotations and documentation for end user
# If you are looking for implementation check return object of each function


@docstring(parameters=[_running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc, _bucket_units_doc,
                       _bucket_end_condition_doc, _end_condition_per_group_doc, _boundary_tick_bucket_doc,
                       _group_by_doc, _groups_to_display_doc])
def compute(*args, **kwargs):
    """
    Generate object that collects aggregations

    See also
    --------
    **COMPUTE** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> c = otp.agg.compute(running=True, bucket_interval=2)
    >>> c.add('X_MEAN', otp.agg.average("X"))
    >>> c.add('X_STD', otp.agg.stddev("X"))
    >>> data = c.apply(data)
    >>> otp.run(data)
                         Time  X_MEAN     X_STD
    0 2003-12-01 00:00:00.000     1.0  0.000000
    1 2003-12-01 00:00:01.000     1.5  0.500000
    2 2003-12-01 00:00:01.500     2.0  0.816497
    3 2003-12-01 00:00:02.000     2.5  0.500000
    4 2003-12-01 00:00:03.000     3.5  0.500000
    5 2003-12-01 00:00:03.500     4.0  0.000000
    6 2003-12-01 00:00:05.000     NaN       NaN

    """
    return Compute(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _time_series_type_doc, _large_ints_doc, _null_int_val_doc, _expect_decimals_doc])
def max(*args, **kwargs):
    """
    Return maximum value of input ``column``

    See also
    --------
    **HIGH** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.max('X')})  # OTdirective: snippet-name: Aggregations.max;
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04       4
    """
    return Max(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _time_series_type_doc, _large_ints_doc, _null_int_val_doc, _expect_decimals_doc])
def min(*args, **kwargs):
    """
    Return minimum value of input ``column``

    See also
    --------
    **LOW** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.min('X')})     # OTdirective: snippet-name: Aggregations.min;
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04       1
    """
    return Min(*args, **kwargs)


@docstring(parameters=[_column_doc, _n_doc, _running_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _keep_timestamp_doc, _selection_doc,
                       _time_series_type_doc])
def high_tick(*args, **kwargs):
    """
    Select ``n`` ticks with the highest values in the ``column`` field

    See also
    --------
    **HIGH_TICK** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> agg = otp.agg.high_tick('X', 2)
    >>> data = agg.apply(data)
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:01.500  3
    1 2003-12-01 00:00:03.000  4
    """
    return HighTick(*args, **kwargs)


@docstring(parameters=[_column_doc, _n_doc, _running_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _keep_timestamp_doc, _selection_doc,
                       _time_series_type_doc])
def low_tick(*args, **kwargs):
    """
    Select ``n`` ticks with the lowest values in the ``column`` field

    See also
    --------
    **LOW_TICK** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> agg = otp.agg.low_tick('X', 2)
    >>> data = agg.apply(data)
    >>> otp.run(data)
                     Time  X
    0 2003-12-01 00:00:00  1
    1 2003-12-01 00:00:01  2
    """
    return LowTick(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _selection_doc, _time_series_type_doc])
def high_time(*args, **kwargs):
    """
    Returns timestamp of tick with highest value of input field

    See also
    --------
    **HIGH_TIME** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 4, 3], offset=[0, 1000, 1500, 3000])
    >>> data = data.agg({'RESULT': otp.agg.high_time(['X'])})    # OTdirective: snippet-name: Aggregations.high time;
    >>> otp.run(data)
            Time                  RESULT
    0 2003-12-04 2003-12-01 00:00:01.500
    """
    return HighTime(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _selection_doc, _time_series_type_doc])
def low_time(*args, **kwargs):
    """
    Returns timestamp of tick with lowest value of input field

    See also
    --------
    **LOW_TIME** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[2, 1, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> data = data.agg({'RESULT': otp.agg.low_time(['X'])})     # OTdirective: snippet-name: Aggregations.low time;
    >>> otp.run(data)
            Time              RESULT
    0 2003-12-04 2003-12-01 00:00:01
    """
    return LowTime(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _large_ints_doc, _null_int_val_doc, _expect_decimals_doc,
                       _skip_tick_if_doc, _time_series_type_doc])
def first(*args, **kwargs):
    """
    Return first value of input field

    See also
    --------
    **FIRST** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> agg = otp.agg.first('X')
    >>> data = agg.apply(data)
    >>> otp.run(data)
            Time  X
    0 2003-12-04  1
    """
    return First(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _large_ints_doc, _null_int_val_doc, _expect_decimals_doc,
                       _skip_tick_if_doc, _time_series_type_doc])
def last(*args, **kwargs):
    """
    Return last value of input field

    See also
    --------
    **LAST** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> agg = otp.agg.last('X')
    >>> data = agg.apply(data)
    >>> otp.run(data)
            Time  X
    0 2003-12-04  4
    """
    return Last(*args, **kwargs)


@docstring(parameters=[_running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _time_series_type_doc])
def first_time(*args, **kwargs):
    """
    Return timestamp of first tick

    See also
    --------
    **FIRST_TIME** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> data = data.agg({'RESULT': otp.agg.first_time()})   # OTdirective: snippet-name: Aggregations.first time;
    >>> otp.run(data)
            Time     RESULT
    0 2003-12-04 2003-12-01
    """
    return FirstTime(*args, **kwargs)


@docstring(parameters=[_running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _time_series_type_doc])
def last_time(*args, **kwargs):
    """
    Return timestamp of last tick

    See also
    --------
    **LAST_TIME** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> data = data.agg({'RESULT': otp.agg.last_time()})   # OTdirective: snippet-name: Aggregations.last time;
    >>> otp.run(data)
            Time              RESULT
    0 2003-12-04 2003-12-01 00:00:03
    """
    return LastTime(*args, **kwargs)


@docstring(parameters=[_running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc, _bucket_units_doc,
                       _bucket_end_condition_doc, _end_condition_per_group_doc, _boundary_tick_bucket_doc,
                       _group_by_doc, _groups_to_display_doc])
def count(*args, **kwargs):
    """
    Returns number of ticks

    See also
    --------
    **NUM_TICKS** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.count()})   # OTdirective: snippet-name: Aggregations.count;
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04       4
    """
    return Count(*args, **kwargs)


_price_doc = param_doc(name='price_column',
                       str_annotation='str or Column',
                       desc='price column for vwap',
                       annotation=Union[str, _Column])

_size_doc = param_doc(name='size_column',
                      str_annotation='str or Column',
                      desc='size column for vwap',
                      annotation=Union[str, _Column])


@docstring(parameters=[_price_doc, _size_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc])
def vwap(*args, **kwargs):
    """
    Returns volume weighted average price

    See also
    --------
    **VWAP** OneTick event processor

    Examples
    --------
    >>> # OTdirective: snippet-name: Aggregations.vwap;
    >>> data = otp.Ticks(P=[1, 2, 3, 4], S=[10, 20, 30, 40])
    >>> data = data.agg({'RESULT': otp.agg.vwap('P','S')})
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04     3.0
    """
    return Vwap(*args, **kwargs)


@docstring(parameters=[_running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc])
def correlation(*args, **kwargs):
    """
    Returns Pearson correlation coefficient value between two numeric fields.

    Parameters
    ----------
    column_name_1: str
        The name of the field in the input stream whose value is to be used by the aggregation
    column_name_2: str
        The name of the field in the input stream whose value is to be used by the aggregation

    See also
    --------
    **CORRELATION** OneTick event processor

    Examples
    --------
    >>> # OTdirective: snippet-name: Aggregations.correlation;
    >>> data = otp.Ticks(P=[1, 2, 3, 4], S=[10, 20, 30, 40])
    >>> data = data.agg({'RESULT': otp.agg.correlation('P','S')})
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04     1.0

    >>> data = otp.Ticks(P=[1, 2, 3, 4], S=[40, 30, 20, 10])
    >>> data = otp.agg.correlation('P', 'S').apply(data)
    >>> otp.run(data)
            Time  CORRELATION
    0 2003-12-04         -1.0

    >>> data = otp.Ticks(P=[1, 2, 3, 4], S=[40, 30, 10, 20])
    >>> data = otp.agg.correlation('P', 'S', bucket_units='ticks', bucket_interval=2).apply(data)
    >>> otp.run(data)
                         Time  CORRELATION
    0 2003-12-01 00:00:00.001         -1.0
    1 2003-12-01 00:00:00.003          1.0
    """
    return Correlation(*args, **kwargs)


@docstring(parameters=[_n_doc, _running_doc, _bucket_interval_doc, _bucket_time_doc, _bucket_units_doc,
                       _bucket_end_condition_doc, _end_condition_per_group_doc, _boundary_tick_bucket_doc,
                       _group_by_doc, _groups_to_display_doc, _keep_timestamp_doc, _default_tick_doc])
def first_tick(*args, **kwargs):
    """
    Select the first **n** ticks

    See also
    --------
    **FIRST_TICK** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(P=[1, 2, 3, 4], S=[10, 20, 30, 40])
    >>> agg = otp.agg.first_tick()
    >>> data = agg.apply(data)
    >>> otp.run(data)
            Time  P   S
    0 2003-12-01  1  10
    """
    return FirstTick(*args, **kwargs)


@docstring(parameters=[_n_doc, _running_doc, _bucket_interval_doc, _bucket_time_doc, _bucket_units_doc,
                       _bucket_end_condition_doc, _end_condition_per_group_doc, _boundary_tick_bucket_doc,
                       _group_by_doc, _groups_to_display_doc, _keep_timestamp_doc, _time_series_type_doc])
def last_tick(*args, **kwargs):
    """
    Select the last ``n`` ticks

    See also
    --------
    **LAST_TICK** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(P=[1, 2, 3, 4], S=[10, 20, 30, 40], offset=[0, 1000, 1500, 3000])
    >>> agg = otp.agg.last_tick()
    >>> data = agg.apply(data)
    >>> otp.run(data)
                     Time  P   S
    0 2003-12-01 00:00:03  4  40
    """
    return LastTick(*args, **kwargs)


@docstring(parameters=[_running_doc, _bucket_interval_doc, _bucket_time_doc, _bucket_units_doc,
                       _bucket_end_condition_doc, _boundary_tick_bucket_doc, _selection_doc])
def distinct(*args, **kwargs):
    """
    Outputs all distinct values for a specified set of key fields.

    Parameters
    ----------
    keys: str or list
        Specifies a list of tick attributes for which unique values are found. The ticks in the input time series
        must contain those attributes.
    key_attrs_only: bool
        If set to true, output ticks will contain only key fields. Otherwise, output ticks will contain all fields
        of an input tick in which a given distinct combination of key values was first encountered.

    See also
    --------
    **DISTINCT** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(dict(x=[1, 3, 1, 5, 3]))
    >>> d = otp.agg.distinct('x')
    >>> data = d.apply(data)
    >>> otp.run(data)
            Time  x
    0 2003-12-04  1
    1 2003-12-04  3
    2 2003-12-04  5
    """
    return Distinct(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc, _expect_decimals_doc])
def sum(*args, **kwargs):
    r"""
    Implement sum aggregation

    See also
    --------
    **SUM** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.sum('X')})  # OTdirective: snippet-name: Aggregations.sum;
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04      10

    Note that **seconds** bucket unit doesn't take into account daylight-saving time of the timezone,
    so you may not get expected results when using, for example, 24 * 60 * 60 seconds as bucket interval.

    >>> data = otp.DataSource('CME', symbols=r'NQ\H23', tick_type='TRD')  # doctest: +SKIP
    >>> data = data.agg({'VOLUME': otp.agg.sum('SIZE')},  # doctest: +SKIP
    ...                 bucket_interval=24*60*60, bucket_units='seconds', bucket_time='start')
    >>> otp.run(data, start=otp.dt(2023, 3, 11), end=otp.dt(2023, 3, 15), timezone='EST5EDT')  # doctest: +SKIP
                     Time  VOLUME
    0 2023-03-11 00:00:00       0
    1 2023-03-12 00:00:00   66190
    2 2023-03-13 01:00:00  631750
    3 2023-03-14 01:00:00  345952

    In such case use **days** bucket unit instead:

    >>> data = otp.DataSource('CME', symbols=r'NQ\H23', tick_type='TRD')  # doctest: +SKIP
    >>> data = data.agg({'VOLUME': otp.agg.sum('SIZE')},  # doctest: +SKIP
    ...                 bucket_interval=1, bucket_units='days', bucket_time='start')
    >>> otp.run(data, start=otp.dt(2023, 3, 11), end=otp.dt(2023, 3, 15), timezone='EST5EDT')  # doctest: +SKIP
            Time  VOLUME
    0 2023-03-11       0
    1 2023-03-12   62940
    2 2023-03-13  634172
    3 2023-03-14  346780
    """
    return Sum(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc])
def average(*args, **kwargs):
    """
    Implement average aggregation

    See also
    --------
    **AVERAGE** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.average('X')})   # OTdirective: snippet-name: Aggregations.average/mean;
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04     2.5
    """
    return Average(*args, **kwargs)


def mean(*args, **kwargs):
    """
    Implement average aggregation.

    See also
    --------
    | :func:`average`
    | **AVERAGE** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.mean('X')})
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04     2.5
    """
    return average(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc, _biased_doc])
def stddev(*args, **kwargs):
    """
    Implement standard deviation aggregation

    See also
    --------
    **STDDEV** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.stddev('X')})    # OTdirective: snippet-name: Aggregations.stddev;
    >>> otp.run(data)
            Time    RESULT
    0 2003-12-04  1.118034
    """
    return StdDev(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc, _time_series_type_w_doc])
def tw_average(*args, **kwargs):
    """
    Returns time weighted average of input field

    See also
    --------
    **TW_AVERAGE** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> otp.run(data, start=otp.dt(2023, 4, 25), end=otp.dt(2023, 4, 25)+otp.Second(4))
                        Time	X
    0	2023-04-25 00:00:00.000	1
    1	2023-04-25 00:00:01.000	2
    2	2023-04-25 00:00:01.500	3
    3	2023-04-25 00:00:03.000	4

    >>> # OTdirective: snippet-name: Aggregations.time weighted average;
    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> data = data.agg({'RESULT': otp.agg.tw_average('X')})
    >>> otp.run(data, start=otp.dt(2023, 4, 25), end=otp.dt(2023, 4, 25)+otp.Second(4))
            Time  RESULT
    0 2023-04-25 00:00:04  2.625
    """
    return TimeWeightedAvg(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc, _expect_decimals_doc])
def median(*args, **kwargs):
    """
    Implement median aggregation

    See also
    --------
    **MEDIAN** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.median('X')})    # OTdirective: snippet-name: Aggregations.median;
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04     3.0
    """
    return Median(*args, **kwargs)


@docstring(parameters=OB_SNAPSHOT_DOC_PARAMS)
def ob_snapshot(*args, **kwargs):
    """
    Returns the order book state at the end of each bucket interval:
    the price, the size, the side, and the time of the last update for a specified number of order book levels.

    See also
    --------
    | :func:`onetick.py.ObSnapshot`
    | **OB_SNAPSHOT** OneTick event processor


    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = otp.agg.ob_snapshot(max_levels=1).apply(data)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  PRICE             UPDATE_TIME  SIZE  LEVEL  BUY_SELL_FLAG
    0 2003-12-04    2.0 2003-12-01 00:00:00.003     6      1              1
    1 2003-12-04    5.0 2003-12-01 00:00:00.004     7      1              0
    """
    return ObSnapshot(*args, **kwargs)


@docstring(parameters=OB_SNAPSHOT_WIDE_DOC_PARAMS)
def ob_snapshot_wide(*args, **kwargs):
    """
    Returns a side by side order book state at the end of each interval:
    the price, the size, and the time of the last update for a specified number of order book levels.

    See also
    --------
    | :func:`onetick.py.ObSnapshotWide`
    | **OB_SNAPSHOT_WIDE** OneTick event processor

    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = otp.agg.ob_snapshot_wide(max_levels=1).apply(data)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  BID_PRICE         BID_UPDATE_TIME  BID_SIZE  ASK_PRICE         ASK_UPDATE_TIME  ASK_SIZE  LEVEL
    0 2003-12-03        5.0 2003-12-01 00:00:00.004         7        2.0 2003-12-01 00:00:00.003         6      1
    """
    return ObSnapshotWide(*args, **kwargs)


@docstring(parameters=OB_SNAPSHOT_FLAT_DOC_PARAMS)
def ob_snapshot_flat(*args, **kwargs):
    """
    Returns the snapshot for a specified number of book levels as a single tick
    with multiple field groups corresponding to book levels.

    See also
    --------
    | :func:`onetick.py.ObSnapshotFlat`
    | **OB_SNAPSHOT_FLAT** OneTick event processor

    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = otp.agg.ob_snapshot_flat(max_levels=1).apply(data)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  BID_PRICE1        BID_UPDATE_TIME1  BID_SIZE1  ASK_PRICE1        ASK_UPDATE_TIME1  ASK_SIZE1
    0 2003-12-03         5.0 2003-12-01 00:00:00.004          7         2.0 2003-12-01 00:00:00.003          6
    """
    return ObSnapshotFlat(*args, **kwargs)


@docstring(parameters=OB_SUMMARY_DOC_PARAMS)
def ob_summary(*args, **kwargs):
    """
    Computes summary statistics, such as:
    VWAP, best and worst price, total size, and number of levels, for one or both sides of an order book.

    See also
    --------
    | :func:`onetick.py.ObSummary`
    | **OB_SUMMARY** OneTick event processor


    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = otp.agg.ob_summary(max_levels=1).apply(data)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  BID_PRICE  BID_SIZE  BID_VWAP  BEST_BID_PRICE  WORST_BID_SIZE  NUM_BID_LEVELS  ASK_SIZE\
              ASK_VWAP  BEST_ASK_PRICE  WORST_ASK_PRICE  NUM_ASK_LEVELS
    0 2003-12-04        NaN         7       5.0             5.0             NaN               1         6\
           2.0             2.0              2.0               1
    """
    return ObSummary(*args, **kwargs)


@docstring(parameters=OB_SIZE_DOC_PARAMS)
def ob_size(*args, **kwargs):
    """
    Returns the total size for a specified number of order book levels at the end of each bucket interval.

    See also
    --------
    | :func:`onetick.py.ObSize`
    | **OB_SIZE** OneTick event processor

    Examples
    --------

    Basic example

    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = otp.agg.ob_size(bucket_interval=otp.Second(300), max_levels=10).apply(data)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
                         Time  ASK_VALUE  BID_VALUE
    0 2003-12-01 00:05:00.000     194800      47200
    1 2003-12-01 00:10:00.000     194800      47200
    2 2003-12-01 00:15:00.000     313400       7700
    ...

    Selecting side via ``side`` parameter

    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = otp.agg.ob_size(bucket_interval=otp.Second(300), max_levels=10, size='ASK').apply(data)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
                         Time   VALUE
    0 2003-12-01 00:05:00.000  194800
    1 2003-12-01 00:10:00.000  194800
    2 2003-12-01 00:15:00.000  313400
    ...
    """
    return ObSize(*args, **kwargs)


@docstring(parameters=OB_VWAP_DOC_PARAMS)
def ob_vwap(*args, **kwargs):
    """
    Returns the size-weighted price computed over a specified number of order book levels at the end of each interval.

    See also
    --------
    | :func:`onetick.py.ObVwap`
    | **OB_VWAP** OneTick event processor

    Examples
    --------

    Basic example

    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = otp.agg.ob_vwap(bucket_interval=otp.Second(300)).apply(data)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
                         Time  ASK_VALUE  BID_VALUE
    0 2003-12-01 00:05:00.000   23.93496   23.80502
    1 2003-12-01 00:10:00.000   23.93496   23.80502
    2 2003-12-01 00:15:00.000   24.35387   24.35387
    ...

    Selecting side via ``side`` parameter

    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = otp.agg.ob_vwap(bucket_interval=otp.Second(300), size='BID').apply(data)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
                         Time     VALUE
    0 2003-12-01 00:05:00.000  23.80502
    1 2003-12-01 00:10:00.000  23.80502
    2 2003-12-01 00:15:00.000  24.35387
    ...
    """
    return ObVwap(*args, **kwargs)


@docstring(parameters=OB_NUM_LEVELS_DOC_PARAMS)
def ob_num_levels(*args, **kwargs):
    """
    Returns the number of levels in the order book at the end of each bucket.

    Note
    ----
    This EP supports only seconds as ``bucket_interval``.

    See also
    --------
    | :func:`onetick.py.ObNumLevels`
    | **OB_NUM_LEVELS** OneTick event processor

    Examples
    --------

    Basic example

    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = otp.agg.ob_num_levels(bucket_interval=otp.Second(300)).apply(data)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
                         Time  ASK_VALUE  BID_VALUE
    0 2003-12-01 00:05:00.000        243         74
    1 2003-12-01 00:10:00.000        243         74
    2 2003-12-01 00:15:00.000        208         45
    ...
    """
    return ObNumLevels(*args, **kwargs)


@docstring(parameters=[
    _query_fun_doc,
    _bucket_delimiter_doc,
    _bucket_interval_doc,
    _bucket_units_doc,
    _bucket_time_doc,
    _bucket_end_condition_doc,
    _running_doc,
    _group_by_doc,
    _groups_to_display_doc,
    _end_condition_per_group_doc,
    _boundary_tick_bucket_doc,
])
def generic(*args, **kwargs):
    """
    Generic aggregation.
    Aggregation logic is provided in ``query_fun`` parameter
    and this logic is applied for ticks in each bucket.
    Currently, this aggregation can be used only with ``.apply()`` method.

    Note, that ``query_fun`` should return a :py:class:`~onetick.py.Source` object,
    assuming that resulted query have only one tick per bucket.

    Also, ``query_fun`` could have additional parameters, which will be passed to ``query_fun``
    during aggregation. Those parameters should be specified in ``.apply()`` as keyword arguments,
    ex: ``.apply(src, additional_param=1)``.

    Generic aggregations could be also used in :py:meth:`onetick.py.Source.agg` method.
    Dict key set for an aggregation in ``aggs`` parameter will be used as prefix
    for each output column of this aggregation.
    Also, all tick fields not contained in the Python-level schema in a passed to generic aggregation
    :class:`Source` object will be removed.

    Note
    ----
    Some functions may be not supported in ``query_fun``.
    For example, :py:func:`~onetick.py.join` and :py:meth:`~onetick.py.Source.rename`.

    See also
    --------
    **GENERIC_AGGREGATION** OneTick event processor

    Examples
    --------

    The simplest case, just copying some other aggregation logic:

    >>> data = otp.Ticks({'A': [1, 2, 3]})
    >>> def agg_fun(source):
    ...     return source.agg({'X': otp.agg.count()})
    >>> data = otp.agg.generic(agg_fun).apply(data)
    >>> otp.run(data)
            Time  X
    0 2003-12-04  3

    Passing parameters to aggregation function:

    >>> data = otp.Ticks({'A': [1, 2, 1]})
    >>> def count_values(source, value):
    ...     values = source.where(source['A'] == value)
    ...     return values.agg({'count': otp.agg.count()})
    >>> data = otp.agg.generic(count_values).apply(data, value=1)
    >>> otp.run(data)
            Time  count
    0 2003-12-04  2

    Getting first 3 ticks from 5 milliseconds buckets:

    >>> data = otp.Ticks({'A': list(range(10))})
    >>> def agg_fun(source, n):
    ...     return source.first(n)
    >>> data = otp.agg.generic(agg_fun, bucket_interval=0.005).apply(data, n=3)
    >>> otp.run(data, start=otp.config.default_start_time, end=otp.config.default_start_time + otp.Milli(10))
                         Time  A
    0 2003-12-01 00:00:00.005  0
    1 2003-12-01 00:00:00.005  1
    2 2003-12-01 00:00:00.005  2
    3 2003-12-01 00:00:00.010  5
    4 2003-12-01 00:00:00.010  6
    5 2003-12-01 00:00:00.010  7

    Using generic aggregation inside :py:meth:`onetick.py.Source.agg` method:

    >>> def agg_fun(source):
    ...     return source.agg({'SUM': otp.agg.sum('X'), 'AVG': otp.agg.average('X')})
    ...
    >>> data = otp.Ticks(X=[1, 2, 3, 4, 5])
    >>> data = data.agg({'X': otp.agg.generic(agg_fun)})
    >>> otp.run(data)
            Time  X.SUM  X.AVG
    0 2003-12-04     15    3.0
    """

    return Generic(*args, **kwargs)


@docstring(parameters=[_running_doc, _bucket_interval_doc, _bucket_time_doc, _bucket_units_doc,
                       _bucket_end_condition_doc, _boundary_tick_bucket_doc])
def option_price(*args, **kwargs):
    """
    This aggregation requires several parameters to compute the option price.
    Those are, OPTION_TYPE, STRIKE_PRICE, EXPIRATION_DATE or DAYS_TILL_EXPIRATION, VOLATILITY, and INTEREST_RATE.
    Each parameter can be specified, either via a symbol parameter with the same name or via a tick field,
    by specifying the name of that field as an EP parameter, as follows.
    Besides, VOLATILITY and INTEREST_RATE can also be specified as parameters. If they are also specified as fields,
    the parameters value are ignored.
    In either case, the OPTION_TYPE value must be set to either CALL or PUT (case insensitive).
    EXPIRATION_DATE is in YYYYMMDD format, a string in case of a symbol parameter and
    an integer in case of a tick attribute.
    Additionally, NUMBER_OF_STEPS should be specified in case of Cox-Ross-Rubinstein method.

    Note
    ----
    This aggregation is used with ``.apply()``, but latest OneTick builds support also the ``.agg()`` method.

    Parameters
    ----------
    volatility: float
        The historical volatility of the asset’s returns.
    interest_rate: float
        The risk-free interest rate.
    compute_model: str
        Allowed values are `BS` and `CRR`.
        Choose between Black–Scholes (`BS`) and Cox-Ross-Rubinstein (`CRR`) models for computing call/put option price.
        Default: `BS`
    number_of_steps: int
        Specifies the number of time steps between the valuation and expiration dates.
        This is a mandatory parameter for `CRR` model.
    compute_delta: bool
        Specifies whether Delta is to be computed or not. This parameter is used only in case of `BS` model.
        Default: False
    compute_gamma: bool
        Specifies whether Gamma is to be computed or not. This parameter is used only in case of `BS` model.
        Default: False
    compute_theta: bool
        Specifies whether Theta is to be computed or not. This parameter is used only in case of `BS` model.
        Default: False
    compute_vega: bool
        Specifies whether Vega is to be computed or not. This parameter is used only in case of `BS` model.
        Default: False
    compute_rho: bool
        Specifies whether Rho is to be computed or not. This parameter is used only in case of `BS` model.
        Default: False
    volatility_field_name: str
        Specifies name of the field, which carries the historical volatility of the asset’s returns.
        Default: empty
    interest_rate_field_name: str
        Specifies name of the field, which carries the risk-free interest rate.
        Default: empty
    option_type_field_name: str
        Specifies name of the field, which carries the option type (either CALL or PUT).
        Default: empty
    strike_price_field_name: str
        Specifies name of the field, which carries the strike price of the option.
        Default: empty
    days_in_year: int
        Specifies number of days in a year (say, 365 or 252 (business days, etc.).
        Used with DAYS_TILL_EXPIRATION parameter to compute the fractional years till expiration.
        Default: 365
    days_till_expiration_field_name: str
        Specifies name of the field, which carries number of days till expiration of the option.
        Default: empty
    expiration_date_field_name: str
        Specifies name of the field, which carries the expiration date of the option, in YYYYMMDD format.
        Default: empty

    See also
    --------
    **OPTION_PRICE** OneTick event processor

    Examples
    ---------
    Black–Scholes with parameters passed through symbol params and calculated delta:

    >>> symbol = otp.Tick(SYMBOL_NAME='SYMB')
    >>> symbol['OPTION_TYPE'] = 'CALL'
    >>> symbol['STRIKE_PRICE'] = 100.0
    >>> symbol['DAYS_TILL_EXPIRATION'] = 30
    >>> symbol['VOLATILITY'] = 0.25
    >>> symbol['INTEREST_RATE'] = 0.05
    >>> data = otp.Ticks(PRICE=[100.7, 101.1, 99.5], symbol=symbol)
    >>> data = otp.agg.option_price(compute_delta=True).apply(data)
    >>> otp.run(data)['SYMB']
            Time     VALUE    DELTA
    0 2003-12-04  2.800999  0.50927
    >>> data.schema
    {'VALUE': <class 'float'>, 'DELTA': <class 'float'>}

    Cox-Ross-Rubinstein with parameters passed through fields:

    >>> data = otp.Ticks(
    ...     PRICE=[100.7, 101.1, 99.5],
    ...     OPTION_TYPE=['CALL']*3,
    ...     STRIKE_PRICE=[100.0]*3,
    ...     DAYS_TILL_EXPIRATION=[30]*3,
    ...     VOLATILITY=[0.25]*3,
    ...     INTEREST_RATE=[0.05]*3,
    ... )
    >>> data = otp.agg.option_price(
    ...     compute_model='CRR',
    ...     number_of_steps=5,
    ...     option_type_field_name='OPTION_TYPE',
    ...     strike_price_field_name='STRIKE_PRICE',
    ...     days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
    ...     volatility_field_name='VOLATILITY',
    ...     interest_rate_field_name='INTEREST_RATE',
    ... ).apply(data)
    >>> otp.run(data)
            Time     VALUE
    0 2003-12-04  2.937537

    Black–Scholes with some parameters passed through parameters:

    >>> data = otp.Ticks(
    ...     PRICE=[100.7, 101.1, 99.5],
    ...     OPTION_TYPE=['CALL']*3,
    ...     STRIKE_PRICE=[100.0]*3,
    ...     DAYS_TILL_EXPIRATION=[30]*3,
    ... )
    >>> data = otp.agg.option_price(
    ...     option_type_field_name='OPTION_TYPE',
    ...     strike_price_field_name='STRIKE_PRICE',
    ...     days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
    ...     volatility=0.25,
    ...     interest_rate=0.05,
    ... ).apply(data)
    >>> otp.run(data)
            Time     VALUE
    0 2003-12-04  2.800999

    To compute values for each tick in a series, set ``bucket_interval=1`` and ``bucket_units='ticks'``

    >>> data = otp.Ticks(
    ...    PRICE=[110.0, 101.0, 112.0],
    ...    OPTION_TYPE=["CALL"]*3,
    ...    STRIKE_PRICE=[110.0]*3,
    ...    DAYS_TILL_EXPIRATION=[30]*3,
    ...    VOLATILITY=[0.2]*3,
    ...    INTEREST_RATE=[0.05]*3
    ... )
    >>> data = otp.agg.option_price(
    ...    option_type_field_name='OPTION_TYPE',
    ...    strike_price_field_name='STRIKE_PRICE',
    ...    days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
    ...    volatility_field_name='VOLATILITY',
    ...    interest_rate_field_name='INTEREST_RATE',
    ...    bucket_interval=1,
    ...    bucket_units='ticks',
    ... ).apply(data)
    >>> otp.run(data)
                           Time 	VALUE
    0 	2003-12-01 00:00:00.000 	2.742714
    1 	2003-12-01 00:00:00.001 	0.212927
    2 	2003-12-01 00:00:00.002 	3.945447

    Usage with the ``.agg()`` method (on the latest OneTick builds).

    .. testcode::
       :skipif: not otp.compatibility.is_supported_agg_option_price()

       data = otp.Ticks(
           PRICE=[100.7, 101.1, 99.5],
           OPTION_TYPE=['CALL']*3,
           STRIKE_PRICE=[100.0]*3,
           DAYS_TILL_EXPIRATION=[30]*3,
       )
       data = data.agg({
           'RESULT': otp.agg.option_price(
               option_type_field_name='OPTION_TYPE',
               strike_price_field_name='STRIKE_PRICE',
               days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
               volatility=0.25,
               interest_rate=0.05,
           )
       })
       df = otp.run(data)
       print(df)

    .. testoutput::

               Time    RESULT
       0 2003-12-04  2.800999

    The following examples show results for different cases of option price calculation.
    Results are compared with two online calculators:
    `Drexel University <https://www.math.drexel.edu/~pg/fin/VanillaCalculator.html>`_ (DU) and
    `Wolfram Alpha <https://www.wolframalpha.com/input?i=black+scholes>`_ (WA).

    Call option, strike price 110.0, underlying price 120.0, volatility 20.0%, interest 5.0%, expiring in 15 days.

    >>> data = {
    ...     "PRICE": 120.,
    ...     "OPTION_TYPE": "call",
    ...     "STRIKE_PRICE": 110.,
    ...     "DAYS_TILL_EXPIRATION": 15,
    ...     "VOLATILITY": 0.2,
    ...     "INTEREST_RATE": 0.05,
    ... }
    >>> data = otp.Tick(**data)
    >>> data = otp.agg.option_price(
    ...     option_type_field_name='OPTION_TYPE',
    ...     strike_price_field_name='STRIKE_PRICE',
    ...     days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
    ...     volatility_field_name='VOLATILITY',
    ...     interest_rate_field_name='INTEREST_RATE',
    ...     compute_delta=True,
    ...     compute_gamma=True,
    ...     compute_theta=True,
    ...     compute_vega=True,
    ...     compute_rho=True
    ... ).apply(data)
    >>> res = otp.run(data).drop("Time", axis=1)
    >>> for key, val in res.to_dict(orient='list').items():
    ...     print(f"{key}={val[0]}")
    VALUE=10.248742578738629
    DELTA=0.9866897658932824
    GAMMA=0.007022082258701294
    THETA=-7.430061156928735
    VEGA=0.8311067221257422
    RHO=4.444686136785831

    .. csv-table:: Benchmark comparison
        :widths: 10, 25, 25, 10
        :header: Field, OneTick, DU benchmark, WA benchmark

        VALUE, 10.248742578738629, 10.248742577611323400, 10.249
        DELTA, 0.9866897658932824, 0.986689766547165200, 0.987
        GAMMA, 0.007022082258701294, 0.007022082258701300, 0.007
        THETA, -7.430061156928735, -7.430061160908399600, -7.430
        VEGA, 0.8311067221257422, 0.831106722125743000, 0.831
        RHO, 4.444686136785831, 4.444686140056785400, 4.445

    Put option, strike price 110.0, underlying price 120.0, volatility 20.0%, interest 5.0%, expiring in 15 days.

    .. testcode::
       :skipif: not otp.compatibility.is_option_price_theta_value_changed()

       data = {
           "PRICE": 120.,
           "OPTION_TYPE": "put",
           "STRIKE_PRICE": 110.,
           "DAYS_TILL_EXPIRATION": 15,
           "VOLATILITY": 0.2,
           "INTEREST_RATE": 0.05,
       }
       data = otp.Tick(**data)
       data = otp.agg.option_price(
           option_type_field_name='OPTION_TYPE',
           strike_price_field_name='STRIKE_PRICE',
           days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
           volatility_field_name='VOLATILITY',
           interest_rate_field_name='INTEREST_RATE',
           compute_delta=True,
           compute_gamma=True,
           compute_theta=True,
           compute_vega=True,
           compute_rho=True
       ).apply(data)
       res = otp.run(data).drop("Time", axis=1)
       for key, val in res.to_dict(orient='list').items():
           print(f"{key}={val[0]:.14f}")

    .. testoutput::

       VALUE=0.02294724243400
       DELTA=-0.01331023410672
       GAMMA=0.00702208225870
       THETA=-1.94135092374397
       VEGA=0.83110672212574
       RHO=-0.06658254802357

    .. csv-table:: Benchmark comparison
        :widths: 10, 25, 25, 10
        :header: Field, OneTick, DU benchmark, WA benchmark

        VALUE, 0.022947242433995818, 0.022947241306682900, 0.023
        DELTA, -0.013310234106717611, -0.013310233452834800, -0.013
        GAMMA, 0.007022082258701294, 0.007022082258701300, 0.007
        THETA, -1.94135092374397, -1.941350927723632000, -1.941
        VEGA, 0.8311067221257422, 0.831106722125743000, 0.831
        RHO, -0.06658254802356636, -0.066582544752611000, -0.067


    Put option, strike price 90.0, underlying price 80.0, volatility 30.0%, interest 8.0%, expiring in 20 days.

    .. testcode::
       :skipif: not otp.compatibility.is_option_price_theta_value_changed()

       data = {
           "PRICE": 80.,
           "OPTION_TYPE": "put",
           "STRIKE_PRICE": 90.,
           "DAYS_TILL_EXPIRATION": 20,
           "VOLATILITY": 0.3,
           "INTEREST_RATE": 0.08,
       }
       data = otp.Tick(**data)
       data = otp.agg.option_price(
           option_type_field_name='OPTION_TYPE',
           strike_price_field_name='STRIKE_PRICE',
           days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
           volatility_field_name='VOLATILITY',
           interest_rate_field_name='INTEREST_RATE',
           compute_delta=True,
           compute_gamma=True,
           compute_theta=True,
           compute_vega=True,
           compute_rho=True
       ).apply(data)
       res = otp.run(data).drop("Time", axis=1)
       for key, val in res.to_dict(orient='list').items():
           print(f"{key}={val[0]}")

    .. testoutput::

       VALUE=9.739720671039635
       DELTA=-0.9429118423759162
       GAMMA=0.020391626464263516
       THETA=0.9410250231811439
       VEGA=2.1453108389800515
       RHO=-4.666995510197969

    .. csv-table:: Benchmark comparison
        :widths: 10, 25, 25, 10
        :header: Field, OneTick, DU benchmark, WA benchmark

        VALUE, 9.739720671039635, 9.739720664487278600, 9.740
        DELTA, -0.9429118423759162, -0.942911845180447200, -0.943
        GAMMA, 0.020391626464263516, 0.020391626464263600, 0.020
        THETA, 0.9410250231811439, 0.941025040605956600, 0.941
        VEGA, 2.1453108389800515, 2.145310838980050700, 2.145
        RHO, -4.666995510197969, -4.666995522132770800, -4.667

    Call option, strike price 90.0, underlying price 80.0, volatility 30.0%, interest 8.0%, expiring in 20 days.

    >>> data = {
    ...     "PRICE": 80.,
    ...     "OPTION_TYPE": "call",
    ...     "STRIKE_PRICE": 90.,
    ...     "DAYS_TILL_EXPIRATION": 20,
    ...     "VOLATILITY": 0.3,
    ...     "INTEREST_RATE": 0.08,
    ... }
    >>> data = otp.Tick(**data)
    >>> data = otp.agg.option_price(
    ...     option_type_field_name='OPTION_TYPE',
    ...     strike_price_field_name='STRIKE_PRICE',
    ...     days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
    ...     volatility_field_name='VOLATILITY',
    ...     interest_rate_field_name='INTEREST_RATE',
    ...     compute_delta=True,
    ...     compute_gamma=True,
    ...     compute_theta=True,
    ...     compute_vega=True,
    ...     compute_rho=True
    ... ).apply(data)
    >>> res = otp.run(data).drop("Time", axis=1)
    >>> for key, val in res.to_dict(orient='list').items():
    ...     print(f"{key}={val[0]:.13f}")
    VALUE=0.1333777785229
    DELTA=0.0570881576241
    GAMMA=0.0203916264643
    THETA=-6.2274824082202
    VEGA=2.1453108389801
    RHO=0.2429410866523

    .. csv-table:: Benchmark comparison
        :widths: 10, 25, 25, 10
        :header: Field, OneTick, DU benchmark, WA benchmark

        VALUE, 0.13337777852292199, 0.133377771970562400, 0.133
        DELTA, 0.05708815762408384, 0.057088154819552600, 0.057
        GAMMA, 0.020391626464263516, 0.020391626464263600, 0.020
        THETA, -6.227482408220195, -6.227482390795381800, -6.227
        VEGA, 2.1453108389800515, 2.145310838980050700, 2.145
        RHO, 0.2429410866522622, 0.242941074717460500, 0.243


    Call option, strike price 140.0, underlying price 150.0, volatility 60.0%, interest 7.0%, expiring in 10 days.

    >>> data = {
    ...     "PRICE": 150.,
    ...     "OPTION_TYPE": "call",
    ...     "STRIKE_PRICE": 140.,
    ...     "DAYS_TILL_EXPIRATION": 10,
    ...     "VOLATILITY": 0.6,
    ...     "INTEREST_RATE": 0.07,
    ... }
    >>> data = otp.Tick(**data)
    >>> data = otp.agg.option_price(
    ...     option_type_field_name='OPTION_TYPE',
    ...     strike_price_field_name='STRIKE_PRICE',
    ...     days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
    ...     volatility_field_name='VOLATILITY',
    ...     interest_rate_field_name='INTEREST_RATE',
    ...     compute_delta=True,
    ...     compute_gamma=True,
    ...     compute_theta=True,
    ...     compute_vega=True,
    ...     compute_rho=True
    ... ).apply(data)
    >>> res = otp.run(data).drop("Time", axis=1)
    >>> for key, val in res.to_dict(orient='list').items():
    ...     print(f"{key}={val[0]:.13f}")
    VALUE=12.2728332229748
    DELTA=0.7774682547236
    GAMMA=0.0200066930955
    THETA=-88.3314253858302
    VEGA=7.3997358024512
    RHO=2.8588330133030

    .. csv-table:: Benchmark comparison
        :widths: 10, 25, 25, 10
        :header: Field, OneTick, DU benchmark, WA benchmark

        VALUE, 12.272833222974782, 12.272833124496244700, 12.27
        DELTA, 0.7774682547235652, 0.777468265655730700, 0.777
        GAMMA, 0.020006693095516285, 0.020006693095516300, 0.020
        THETA, -88.33142538583016, -88.331425507511386000, -88.331
        VEGA, 7.399735802451228, 7.399735802451227600, 7.400
        RHO, 2.858833013303013, 2.858833060927763500, 2.859


    Put option, strike price 140.0, underlying price 150.0, volatility 60.0%, interest 7.0%, expiring in 10 days.

    .. testcode::
       :skipif: not otp.compatibility.is_option_price_theta_value_changed()

       data = {
           "PRICE": 150.,
           "OPTION_TYPE": "put",
           "STRIKE_PRICE": 140.,
           "DAYS_TILL_EXPIRATION": 10,
           "VOLATILITY": 0.6,
           "INTEREST_RATE": 0.07,
       }
       data = otp.Tick(**data)
       data = otp.agg.option_price(
           option_type_field_name='OPTION_TYPE',
           strike_price_field_name='STRIKE_PRICE',
           days_till_expiration_field_name='DAYS_TILL_EXPIRATION',
           volatility_field_name='VOLATILITY',
           interest_rate_field_name='INTEREST_RATE',
           compute_delta=True,
           compute_gamma=True,
           compute_theta=True,
           compute_vega=True,
           compute_rho=True).apply(data)
       res = otp.run(data).drop("Time", axis=1)
       for key, val in res.to_dict(orient='list').items():
           print(f"{key}={val[0]:.13f}")

    .. testoutput::

       VALUE=2.0045973669685
       DELTA=-0.2225317452764
       GAMMA=0.0200066930955
       THETA=-78.5502018957506
       VEGA=7.3997358024512
       RHO=-0.9694344974913

    .. csv-table:: Benchmark comparison
        :widths: 10, 25, 25, 10
        :header: Field, OneTick, DU benchmark, WA benchmark

        VALUE, 2.0045973669685395, 2.004597268490016500, 2.00
        DELTA, -0.22253174527643482, -0.222531734344269000, -0.223
        GAMMA, 0.020006693095516285, 0.020006693095516300, 0.020
        THETA, -78.5502018957506, -78.550202017431814100, -78.550
        VEGA, 7.399735802451228, 7.399735802451227600, 7.400
        RHO, -0.9694344974913363, -0.969434449866586000, -0.969

    """

    return OptionPrice(*args, **kwargs)


@docstring(parameters=[
    _bucket_interval_doc,
    _bucket_units_doc,
    _bucket_time_doc,
    _bucket_end_condition_doc,
    _boundary_tick_bucket_doc,
    _group_by_doc,
    _groups_to_display_doc,
    _end_condition_per_group_doc,
])
def ranking(*args, **kwargs):
    """
    Ranking **running** aggregation.

    Sorts a series of ticks over a bucket interval
    using a specified set of tick fields specified in ``rank_by``
    and adds a new field ``RANKING``
    with the position of the tick in the sort order
    or the percentage of ticks with values less than (or equal) to the value of the tick.

    Does not change the order of the ticks.

    See also
    --------
    **RANKING** OneTick event processor

    Parameters
    ----------
    rank_by: str or list or dict
        Set of fields to sort by.
        Can be one field specified by string, list of fields or dictionary with field names
        as keys and ``asc`` or ``desc`` string literals as values. Latter allows to specify
        sorting direction. Default direction is ``desc``.
    show_rank_as: str

        - ``order``: calculate number that represents the position of the tick in the sort order
        - ``percent_le_values``: calculate the percentage of ticks that have higher or equal value\
           of the position in the sort order, relative to the tick
        - ``percent_lt_values``: calculate the percentage of ticks that have higher value\
           of the position in the sort order, relative to the tick
        - ``percentile_standard``: calculate Percentile Rank of the tick in the sort order.
    include_tick: bool, default=False
        Specifies whether the current tick should be included in calculations
        if ``show_rank_as`` is ``percent_lt_values`` or ``percentile_standard``.

    Examples
    --------
    >>> t = otp.Ticks({'A': [1, 2, 3]})
    >>> t = t.ranking('A')
    >>> otp.run(t)
                         Time  A  RANKING
    0 2003-12-01 00:00:00.000  1        3
    1 2003-12-01 00:00:00.001  2        2
    2 2003-12-01 00:00:00.002  3        1

    >>> t = otp.Ticks({'A': [1, 2, 3]})
    >>> t = t.ranking({'A': 'asc'})
    >>> otp.run(t)
                         Time  A  RANKING
    0 2003-12-01 00:00:00.000  1        1
    1 2003-12-01 00:00:00.001  2        2
    2 2003-12-01 00:00:00.002  3        3

    >>> t = otp.Ticks({'A': [1, 2, 2, 3, 2, 1]})
    >>> otp.run(t.ranking({'A': 'asc'}, show_rank_as='percent_lt_values', include_tick=True))
                         Time  A    RANKING
    0 2003-12-01 00:00:00.000  1  66.666667
    1 2003-12-01 00:00:00.001  2  16.666667
    2 2003-12-01 00:00:00.002  2  16.666667
    3 2003-12-01 00:00:00.003  3   0.000000
    4 2003-12-01 00:00:00.004  2  16.666667
    5 2003-12-01 00:00:00.005  1  66.666667

    >>> otp.run(t.ranking({'A': 'asc'}, show_rank_as='percent_lt_values', include_tick=False))
                         Time  A  RANKING
    0 2003-12-01 00:00:00.000  1     80.0
    1 2003-12-01 00:00:00.001  2     20.0
    2 2003-12-01 00:00:00.002  2     20.0
    3 2003-12-01 00:00:00.003  3      0.0
    4 2003-12-01 00:00:00.004  2     20.0
    5 2003-12-01 00:00:00.005  1     80.0
    """
    return Ranking(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc])
def variance(*args, **kwargs):
    """
    Implement variance aggregation

    Parameters
    ----------
    biased: bool
        Switches between biased and unbiased variance calculation.

    See also
    --------
    **VARIANCE** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.variance('X', biased=False)})
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04     1.3

    >>> data = otp.Ticks(X=[1, 2, 3, 3, 4])
    >>> data = data.agg({'RESULT': otp.agg.variance('X', biased=True)})
    >>> otp.run(data)
            Time  RESULT
    0 2003-12-04     1.04
    """
    return Variance(*args, **kwargs)


@docstring(parameters=[
    _running_doc,
    _bucket_interval_doc,
    _bucket_units_doc,
    _bucket_time_doc,
    _bucket_end_condition_doc,
    _boundary_tick_bucket_doc,
    _group_by_doc,
    _groups_to_display_doc,
    _end_condition_per_group_doc,
])
def percentile(*args, **kwargs):
    """
    Percentile **running** aggregation.

    For each bucket, propagates its ``n-1`` ``n-quantiles`` where a comparison between ticks is done
    using a specified set of tick fields. A new field (``QUANTILE``) with the quantile number is added.

    See also
    --------
    **PERCENTILE** OneTick event processor

    Parameters
    ----------
    number_of_quantiles: int
        Specifies the number ``n`` of quantiles. Setting it to 2 will propagate only one tick - the median.

        Default: 2
    input_field_names: List[Union[Union[str, onetick.py.Column], Tuple[Union[str, onetick.py.Column], str]]]
        List of numeric field names to run aggregation on.
        You can use as list elements either string column name, either :py:class:`Columns <onetick.py.Column>`.

        You can change default comparison order (``desc``) for the field,
        by passing a tuple of column name and comparison order (``desc`` or ``asc``) instead column name.
    output_field_names: Optional[List[str]]
        Output columns name in the same order as columns from ``input_field_names``

        ``output_field_names`` and ``input_field_names`` must have the same number of fields.

        If not set, ``output_field_names`` will be equal to ``input_field_names``.

    Examples
    --------
    >>> t = otp.Ticks({'A': [1, 2, 3, 4]})
    >>> t = t.percentile(['A'])
    >>> otp.run(t)
            Time  A  QUANTILE
    0 2003-12-04  2         1

    You can also pass column:

    >>> t = otp.Ticks({'A': [1, 2, 3, 4]})
    >>> t = t.percentile([t['A']])
    >>> otp.run(t)
            Time  A  QUANTILE
    0 2003-12-04  2         1

    Change number of quantiles:

    >>> t = otp.Ticks({'A': [1, 2, 3, 4]})
    >>> t = t.percentile(['A'], number_of_quantiles=3)
    >>> otp.run(t)
            Time  A  QUANTILE
    0 2003-12-04  3         1
    1 2003-12-04  2         2

    Or change default comparison order:

    >>> t = otp.Ticks({'A': [1, 2, 3, 4]})
    >>> t = t.percentile([('A', 'asc')], number_of_quantiles=3)
    >>> otp.run(t)
            Time  A  QUANTILE
    0 2003-12-04  2         1
    1 2003-12-04  3         2

    You can also change output column name via ``output_field_names`` parameter:

    >>> t = otp.Ticks({'A': [1, 2, 3, 4]})
    >>> t = t.percentile(['A'], output_field_names=['B'])
    >>> otp.run(t)
            Time  B  QUANTILE
    0 2003-12-04  2         1
    """
    return Percentile(*args, **kwargs)


@docstring(parameters=[
    _running_doc,
    _bucket_interval_doc,
    _bucket_units_doc,
    _bucket_time_doc,
    _bucket_end_condition_doc,
    _boundary_tick_bucket_doc,
    _group_by_doc,
    _groups_to_display_doc,
    _end_condition_per_group_doc,
])
def find_value_for_percentile(*args, **kwargs):
    """
    For each bucket, finds the value which has percentile rank closest to ``percentile`` parameter.

    For more details, see `Percentile rank <https://en.wikipedia.org/wiki/Percentile_rank>`_.

    See also
    --------
    **FIND_VALUE_FOR_PERCENTILE** OneTick event processor

    Parameters
    ----------
    percentile: int
        Specifies the percentile number in 0-100% range
        for which find the value of input field that has percentile rank close to it.
    show_percentile_as: str
        * By default (when parameter is not set),
          the output field contains the value which has percentile rank closest to ``percentile`` parameter.
        * When set to *interpolated_value*, the output field contains the interpolated value of:
          The value above and below specified ``percentile``, or one of them if it's biggest or smallest value.
        * When set to *first_value_with_ge_percentile*, the output field contains
          the value which has greater or equal than specified ``percentile``.

    Examples
    --------

    Find interpolated value of 50% percentile rank:

    .. testcode::
       :skipif: not otp.compatibility.is_find_value_for_percentile_supported()

       t = otp.Ticks({'A': [1, 2, 3, 4]})
       data = t.find_value_for_percentile('A', 50, 'interpolated_value')
       df = otp.run(data)
       print(df)

    .. testoutput::

               Time    A
       0 2003-12-04  2.5

    Find first value greater or equal to 50% percentile rank:

    .. testcode::
       :skipif: not otp.compatibility.is_find_value_for_percentile_supported()

       data = t.find_value_for_percentile('A', 50, 'first_value_with_ge_percentile')
       df = otp.run(data)
       print(df)

    .. testoutput::

               Time    A
       0 2003-12-04  3.0

    Use :py:meth:`~onetick.py.Source.agg` method to apply several aggregations at the same time:

    .. testcode::
       :skipif: not otp.compatibility.is_find_value_for_percentile_supported()

       data = t.agg({
           'X': otp.agg.find_value_for_percentile('A', 50, 'interpolated_value'),
           'COUNT': otp.agg.count(),
       })
       df = otp.run(data)
       print(df)

    .. testoutput::

               Time    X  COUNT
       0 2003-12-04  2.5      4
    """
    return FindValueForPercentile(*args, **kwargs)


@docstring(parameters=[
    _decay_doc,
    _decay_value_type_doc,
    _running_doc,
    _bucket_interval_doc,
    _bucket_units_doc,
    _bucket_time_doc,
    _bucket_end_condition_doc,
    _boundary_tick_bucket_doc,
    _all_fields_doc,
    _group_by_doc,
    _groups_to_display_doc,
    _end_condition_per_group_doc,
    _time_series_type_w_doc,
])
def exp_w_average(*args, **kwargs):
    """
    ``EXP_W_AVERAGE`` aggregation.

    For each bucket, computes the **exponentially weighted average** value of the specified numeric attribute.
    Weights of data points in a bucket decrease exponentially in the direction from the most recent tick
    to the most aged one, being equal to ``exp(-Lambda * N)`` for a fixed weight decay value **Lambda**,
    where **N** ranges over **0, 1, 2, …** as ticks in reverse order of their arrival are treated.
    Once the weights are known, the average is found using the formula ``sum(weight*value)/sum(weight)``,
    where the sum is computed across all data points.

    See also
    --------
    **EXP_W_AVERAGE** OneTick event processor

    Examples
    --------

    Basic example

    >>> data = otp.Ticks({'A': [1.0, 2.0, 3.0, 3.0, 4.0]})
    >>> data = data.exp_w_average('A', decay=2, bucket_interval=2, bucket_units='ticks')
    >>> otp.run(data)
                         Time         A
    0 2003-12-01 00:00:00.001  1.880797
    1 2003-12-01 00:00:00.003  2.984124
    2 2003-12-04 00:00:00.000  3.880797

    You can switch to ``half_life_index`` as ``decay_value_type``

    >>> data = otp.Ticks({'A': [1.0, 2.0, 3.0, 3.0, 4.0]})
    >>> data = data.exp_w_average(
    ...     'A', decay=2, decay_value_type='half_life_index', bucket_interval=2, bucket_units='ticks',
    ... )
    >>> otp.run(data)
                         Time         A
    0 2003-12-01 00:00:00.001  1.585786
    1 2003-12-01 00:00:00.003  2.773459
    2 2003-12-04 00:00:00.000  3.585786
    """
    return ExpWAverage(*args, **kwargs)


@docstring(parameters=[
    _decay_doc,
    _decay_value_type_hl_doc,
    _running_doc,
    _bucket_interval_doc,
    _bucket_units_doc,
    _bucket_time_doc,
    _bucket_end_condition_doc,
    _boundary_tick_bucket_doc,
    _all_fields_doc,
    _group_by_doc,
    _groups_to_display_doc,
    _end_condition_per_group_doc,
])
def exp_tw_average(*args, **kwargs):
    """
    ``EXP_TW_AVERAGE`` aggregation.

    For each bucket, computes the **exponentially time-weighted average** value of a specified numeric field.
    The weight of each point in the time series is computed relative to the end time of the bucket,
    so that the value which is in effect during some infinitely small time interval `delta t`
    has weight **(delta t)*exp(-Lambda*(end_time - t))**,
    where `Lambda` is a constant, `end_time` represents end time of the bucket,
    and `t` represents the timestamp of that infinitely small time interval.

    See also
    --------
    **EXP_TW_AVERAGE** OneTick event processor

    Examples
    --------

    Basic example

    >>> data = otp.Ticks({'A': [1.0, 2.0, 3.0, 3.0, 4.0]})
    >>> data = data.exp_tw_average('A', decay=2, bucket_interval=2, bucket_units='ticks')
    >>> otp.run(data)
                         Time         A
    0 2003-12-01 00:00:00.001  1.000000
    1 2003-12-01 00:00:00.003  2.500087
    2 2003-12-04 00:00:00.000  4.000000

    You can switch to ``lambda`` as ``decay_value_type``

    >>> data = otp.Ticks({'A': [1.0, 2.0, 3.0, 3.0, 4.0]})
    >>> data = data.exp_tw_average(
    ...     'A', decay=2, decay_value_type='lambda', bucket_interval=2, bucket_units='ticks',
    ... )
    >>> otp.run(data)
                         Time       A
    0 2003-12-01 00:00:00.001  1.0000
    1 2003-12-01 00:00:00.003  2.5005
    2 2003-12-04 00:00:00.000  4.0000
    """
    return ExpTwAverage(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc, _degree_doc])
def standardized_moment(*args, **kwargs):
    """
    ``STANDARDIZED_MOMENT`` aggregation.

    Computes the standardized statistical moment of degree **k** of the input ``column``
    over the specified bucket interval.
    The standardized moment of degree **k** is defined
    as the expected value of the expression ``((X - mean) / stddev)^k``.

    See also
    --------
    **STANDARDIZED_MOMENT** OneTick event processor

    Examples
    --------

    Basic example

    .. testcode::
       :skipif: not otp.compatibility.is_standardized_moment_supported()

       data = otp.Ticks({'A': [1, 2, 4, 4, 4, 6]})
       data = data.standardized_moment('A', degree=3, bucket_interval=3, bucket_units='ticks')
       df = otp.run(data)
       print(df)

    .. testoutput::
                            Time         A
       0 2003-12-01 00:00:00.002  0.381802
       1 2003-12-01 00:00:00.005  0.707107
    """
    return StandardizedMoment(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _weight_field_name_doc, _portfolio_side_doc,
                       _weight_type_doc, _symbols_doc])
def portfolio_price(*args, **kwargs):
    """
    ``PORTFOLIO_PRICE`` aggregation.

    For each bucket, computes weighted portfolio price.

    See also
    --------
    **PORTFOLIO_PRICE** OneTick event processor

    Examples
    --------

    Basic example, by default this EP takes ``PRICE`` column as input

    >>> data = otp.DataSource(
    ...     'US_COMP', symbol='AAPL', tick_type='TRD',
    ...     start=otp.dt(2022, 3, 1), end=otp.dt(2022, 3, 2),
    ... )
    >>> data = data.portfolio_price()
    >>> otp.run(data)
            Time  VALUE  NUM_SYMBOLS
    0 2022-03-02    1.4            1

    Getting portfolio price for multiple symbols:

    >>> data = otp.DataSource('US_COMP', tick_type='TRD', start=otp.dt(2022, 3, 1), end=otp.dt(2022, 3, 2))
    >>> data = data.portfolio_price(symbols=['AAPL', 'AAP'])
    >>> otp.run(data)
            Time  VALUE  NUM_SYMBOLS
    0 2022-03-02  46.81            2

    Applying **PORTFOLIO_PRICE** on custom column

    >>> data = otp.Ticks(X=[10.0, 12.5, 11.0, 10.2, 15])
    >>> data = data.portfolio_price('X')
    >>> otp.run(data)
            Time  VALUE  NUM_SYMBOLS
    0 2003-12-04   15.0            1

    Specifying weights via column 'WEIGHS'

    >>> data = otp.Ticks(PRICE=[10.0, 12.5, 11.0, 10.2, 15], WEIGHTS=[1, 2, -1, 2, 2])
    >>> data = data.portfolio_price(weight_field_name=data['WEIGHTS'])
    >>> otp.run(data)
            Time  VALUE  NUM_SYMBOLS
    0 2003-12-04   30.0            1
    """
    return PortfolioPrice(*args, **kwargs)


@docstring(parameters=[_portfolios_query_doc, _columns_portfolio_doc, _running_doc, _bucket_interval_doc,
                       _bucket_time_doc, _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _weight_field_name_doc,
                       _weight_multiplier_field_name_doc, _portfolio_side_doc, _weight_type_doc,
                       _portfolios_query_params_doc, _portfolio_value_field_name_doc, _symbols_doc])
def multi_portfolio_price(*args, **kwargs):
    """
    ``MULTI_PORTFOLIO_PRICE`` aggregation.

    For each bucket, computes weighted portfolio price for multiple portfolios.

    See also
    --------
    **MULTI_PORTFOLIO_PRICE** OneTick event processor

    Examples
    --------

    Basic example, by default this EP takes ``PRICE`` column as input

    >>> data = otp.DataSource(
    ...     'US_COMP', tick_type='TRD', date=otp.dt(2022, 3, 1)
    ... )
    >>> data = data.multi_portfolio_price(
    ...     portfolios_query='some_query.otq::portfolios_query',
    ...     symbols=['US_COMP::AAPL', 'US_COMP::MSFT', 'US_COMP::ORCL'],
    ... )
    >>> otp.run(data)  # doctest: +SKIP
            Time  VALUE  NUM_SYMBOLS PORTFOLIO_NAME
    0 2003-12-01   95.0            3    PORTFOLIO_1
    1 2003-12-01   47.5            1    PORTFOLIO_2
    2 2003-12-01   32.5            2    PORTFOLIO_3

    Override ``weight`` returned by ``portfolios_query`` with ``weight_field_name``

    >>> data = otp.DataSource(
    ...     'US_COMP', tick_type='TRD', date=otp.dt(2022, 3, 1)
    ... )
    >>> data['WEIGHT'] = 2
    >>> data = data.multi_portfolio_price(
    ...     portfolios_query='some_query.otq::portfolios_query',
    ...     weight_field_name='WEIGHT',
    ...     symbols=['US_COMP::AAPL', 'US_COMP::MSFT', 'US_COMP::ORCL'],
    ... )
    >>> otp.run(data)  # doctest: +SKIP
            Time  VALUE  NUM_SYMBOLS PORTFOLIO_NAME
    0 2003-12-01   38.0            3    PORTFOLIO_1
    1 2003-12-01   19.0            1    PORTFOLIO_2
    2 2003-12-01   13.0            2    PORTFOLIO_3

    Pass parameters to the query from ``portfolios_query`` via ``portfolios_query_params``

    >>> data = otp.DataSource(
    ...     'US_COMP', tick_type='TRD', date=otp.dt(2022, 3, 1)
    ... )
    >>> data = data.multi_portfolio_price(
    ...     portfolios_query='some_query.otq::portfolios_query_with_param',
    ...     symbols=['US_COMP::AAPL', 'US_COMP::MSFT', 'US_COMP::ORCL'],
    ...     portfolios_query_params={'PORTFOLIO_1_NAME': 'CUSTOM_NAME'}
    ... )
    >>> otp.run(data)  # doctest: +SKIP
            Time  VALUE  NUM_SYMBOLS PORTFOLIO_NAME
    0 2003-12-01   95.0            3    CUSTOM_NAME
    1 2003-12-01   47.5            1    PORTFOLIO_2
    2 2003-12-01   32.5            2    PORTFOLIO_3

    Use ``otp.Source`` object as ``portfolios_query`` (only for local queries)

    >>> portfolios = otp.Ticks(
    ...     SYMBOL_NAME=['US_COMP::AAPL', 'US_COMP::MSFT', 'US_COMP::AAPL'],
    ...     PORTFOLIO_NAME=['PORTFOLIO_1', 'PORTFOLIO_1', 'PORTFOLIO_2'],
    ...     WEIGHT=[1, 1, 2],
    ... )
    >>> data = otp.DataSource(
    ...     'US_COMP', tick_type='TRD', date=otp.dt(2022, 3, 1)
    ... )
    >>> data = data.multi_portfolio_price(
    ...     portfolios_query=portfolios,
    ...     symbols=['US_COMP::AAPL', 'US_COMP::MSFT'],
    ... )
    >>> otp.run(data)  # doctest: +SKIP
            Time  VALUE  NUM_SYMBOLS PORTFOLIO_NAME
    0 2003-12-01   47.5            2    PORTFOLIO_1
    1 2003-12-01   46.0            1    PORTFOLIO_2
    """
    return MultiPortfolioPrice(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc])
def return_ep(*args, **kwargs):
    """
    ``RETURN`` aggregation.

    Computes the ratio between the value of the input field at the end of a bucket interval and the value
    of this field at the start of a bucket interval.
    If ``running`` is set to ``True``, then ``RETURN`` is calculated as the ratio between latest tick
    and the first tick from the start time. Below, the input field is referred to as 'price'.

    ``Return for an interval = ending price for the interval / starting price for the interval``

    where

    * ``ending price for the interval`` is taken from the last tick observed in the stream
      up to the end of the interval.
    * ``starting price for the interval`` is taken from the first tick in the interval for the first bucket interval
      that has any ticks. For all other intervals, it is the price of the first tick with timestamp of the start
      of interval or, if no such tick is present, the ending price of the previous interval. The ending price
      for the interval is taken from the last tick observed in the stream up to the end of the interval.

    See also
    --------
    **RETURN** OneTick event processor

    Examples
    --------

    Basic example as :py:class:`~onetick.py.Source` method

    >>> data = otp.DataSource('US_COMP', symbol='AAPL', tick_type='TRD')
    >>> data = data.return_ep(data['PRICE'], bucket_interval=otp.Minute(10))
    >>> otp.run(data, date=otp.dt(2022, 3, 1))  # doctest: +SKIP
                         Time    PRICE
    0 2022-03-01 00:00:00.000  0.99953
    1 2022-03-01 00:10:00.000   1.0043
    2 2022-03-01 00:20:00.000   0.9986
    3 2022-03-01 00:30:00.000  0.99643
    4 2022-03-01 00:40:00.000    1.042
    ...

    Basic example as aggregation

    >>> data = otp.DataSource('US_COMP', symbol='AAPL', tick_type='TRD')
    >>> data = otp.agg.return_ep(data['PRICE'], bucket_interval=otp.Minute(10)).apply(data)
    >>> otp.run(data, date=otp.dt(2022, 3, 1))  # doctest: +SKIP
                         Time    PRICE
    0 2022-03-01 00:00:00.000  0.99953
    1 2022-03-01 00:10:00.000   1.0043
    2 2022-03-01 00:20:00.000   0.9986
    3 2022-03-01 00:30:00.000  0.99643
    4 2022-03-01 00:40:00.000    1.042
    ...
    """
    return Return(*args, **kwargs)


@docstring(parameters=[_column_doc, _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc,
                       _interest_rate_doc, _price_field_doc, _option_price_field_doc,
                       _method_doc, _precision_doc, _value_for_non_converge_doc, _option_type_field_doc,
                       _strike_price_field_doc, _days_in_year_doc, _days_till_expiration_field_doc,
                       _expiration_date_field_doc,
                       ])
def implied_vol(*args, **kwargs):
    """
    ``IMPLIED_VOL`` aggregation.

    For each bucket, computes implied volatility value for the last tick in the bucket,
    based on the Black-Scholes option pricing model.

    This EP requires a time series of ticks, having the ``PRICE`` and ``OPTION_PRICE`` attributes.

    It also requires several parameters to compute the implied volatility.
    Those are, ``OPTION_TYPE``, ``STRIKE_PRICE``, ``EXPIRATION_DATE`` or ``DAYS_TILL_EXPIRATION`` and ``INTEREST_RATE``.
    Each parameter can be specified either via a symbol parameter with the same name, or via a tick field,
    by specifying name of that field as an EP parameter.
    Besides, ``interest_rate`` can also be specified as aggregation parameter.
    In either case ``OPTION_TYPE`` must have either ``CALL`` value, or ``PUT``.
    ``EXPIRATION_DATE`` is in ``YYYYMMDD`` format, a string in case of a symbol parameter and an integer
    in case of a tick attribute.

    See also
    --------
    **IMPLIED_VOL** OneTick event processor

    Examples
    --------

    Basic example:

    >>> data = otp.DataSource('SOME_DB', symbol='AAA', tick_type='TT')  # doctest: +SKIP
    >>> data = data.implied_vol(
    ...     interest_rate=0.05, option_type_field=data['OPTION_TYPE'],
    ...     strike_price_field=data['STRIKE_PRICE'], days_till_expiration_field=data['DAYS_TILL_EXPIRATION'],
    ... )  # doctest: +SKIP
    >>> otp.run(data)  # doctest: +SKIP
            Time     VALUE
    0 2003-12-04  0.889491

    Specifying ``interest_rate`` and ``strike_price`` as symbol parameters:

    >>> sym = otp.Ticks({
    ...     'SYMBOL_NAME': ['TEST'],
    ...     'INTEREST_RATE': [0.05],
    ...     'STRIKE_PRICE': [100.0],
    ... })  # doctest: +SKIP
    >>> data = otp.DataSource('SOME_DB', symbol='AAA', tick_type='TT')  # doctest: +SKIP
    >>> data = data.implied_vol(
    ...     option_type_field=data['OPTION_TYPE'], days_till_expiration_field=data['DAYS_TILL_EXPIRATION'],
    ... )  # doctest: +SKIP
    >>> otp.run(data)  # doctest: +SKIP
            Time     VALUE
    0 2003-12-04  0.889491
    """
    return ImpliedVol(*args, **kwargs)


@docstring(parameters=[_dependent_variable_field_name_doc, _independent_variable_field_name_doc,
                       _running_doc, _all_fields_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _end_condition_per_group_doc,
                       _boundary_tick_bucket_doc,
                       ])
def linear_regression(*args, **kwargs):
    """
    ``LINEAR_REGRESSION`` aggregation.

    For each bucket, computes the linear regression parameters slope and intercept of specified input fields
    ``dependent_variable_field_name`` and ``independent_variable_field_name``.
    Adds computed parameters as `SLOPE` and `INTERCEPT` fields in output time series.
    The relationship between the dependent variable (``Y``) and the independent variable (``X``) is defined
    by the formula: Y = SLOPE * X + INTERCEPT, where `SLOPE` and `INTERCEPT` are the calculated output parameters.

    See also
    --------
    **LINEAR_REGRESSION** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks({'X': [10.0, 9.5, 8.0, 8.5], 'Y': [3.0, 5.0, 4.5, 3.5]})
    >>> data = data.linear_regression(
    ...     dependent_variable_field_name=data['Y'],
    ...     independent_variable_field_name=data['X'],
    ... )  # doctest: +SKIP
    >>> otp.run(data)  # doctest: +SKIP
            Time  SLOPE  INTERCEPT
    0 2003-12-04   -0.3        6.7
    """
    return LinearRegression(*args, **kwargs)


@docstring(parameters=[_field_to_partition_doc, _weight_field_doc, _number_of_groups_doc,
                       _running_doc, _bucket_interval_doc, _bucket_time_doc,
                       _bucket_units_doc, _bucket_end_condition_doc, _boundary_tick_bucket_doc,
                       ])
def partition_evenly_into_groups(*args, **kwargs):
    """
    ``PARTITION_EVENLY_INTO_GROUPS`` aggregation.

     For each bucket, this EP breaks ticks into the specified number of groups (``number_of_groups``)
     by the specified field (``field_to_partition``) in a way that the sums
     of the specified weight fields (``weight_field``) in each group are as close as possible.

    See also
    --------
    **PARTITION_EVENLY_INTO_GROUPS** OneTick event processor

    Examples
    --------

    >>> data = otp.Ticks(X=['A', 'B', 'A', 'C', 'D'], SIZE=[10, 30, 20, 15, 14])
    >>> data = data.partition_evenly_into_groups(
    ...     field_to_partition=data['X'],
    ...     weight_field=data['SIZE'],
    ...     number_of_groups=3,
    ... )
    >>> otp.run(data)
            Time FIELD_TO_PARTITION  GROUP_ID
    0 2003-12-04                  A         0
    1 2003-12-04                  B         1
    2 2003-12-04                  C         2
    3 2003-12-04                  D         2
    """
    return PartitionEvenlyIntoGroups(*args, **kwargs)



import os
import sys
import json
import logging
import warnings
from pathlib import Path
from .configuration import config


class CustomStreamHandler(logging.StreamHandler):
    # workaround for https://github.com/pytest-dev/pytest/issues/5502
    # pytest doesn't work well when you initialize logging at import time
    # and then try logging to stdout at application exit
    # (we do it when deleting temporary files)
    # in this case logging system writes very cumbersome error messages,
    # so here we override them
    def handleError(self, record):
        if logging.raiseExceptions and sys.stderr:
            try:
                msg = self.format(record)
                sys.stderr.write(msg + self.terminator)
                sys.stderr.flush()
            except Exception:
                pass


def get_stream_handler():
    handler = CustomStreamHandler()
    formatter = logging.Formatter('[%(asctime)s] %(levelname)s (%(name)s): %(message)s')
    handler.setFormatter(formatter)
    return handler


ROOT_LOGGER_NAME = 'onetick.py'
ROOT_LOGGER = None


def __init_root_logger():
    global ROOT_LOGGER

    if ROOT_LOGGER is not None:
        return

    ROOT_LOGGER = logging.getLogger(ROOT_LOGGER_NAME)
    ROOT_LOGGER.addHandler(logging.NullHandler())

    if not config.logging:
        return

    if os.sep in config.logging:
        config_file = Path(config.logging)
        if not config_file.exists():
            warnings.warn(f"Configuration file '{config_file}' doesn't exist, can't initialize logging system")
            return

        # Config file is crafted and loaded by user, so we can ignore SonarQube warnings
        if config_file.suffix == '.json':
            with config_file.open() as f:
                json_data = json.load(f)
            logging.config.dictConfig(json_data)  # NOSONAR
        else:
            logging.config.fileConfig(config_file)  # NOSONAR
    else:
        try:
            ROOT_LOGGER.setLevel(config.logging)
        except ValueError:
            warnings.warn(f"Unknown logging level name: '{config.logging}', can't initialize logging system")
            return
        ROOT_LOGGER.addHandler(get_stream_handler())

    ROOT_LOGGER.debug(f'Initialized {ROOT_LOGGER_NAME} logging system')


__init_root_logger()


def get_logger(*parts):
    name = '.'.join(parts)
    if not name.startswith(ROOT_LOGGER_NAME):
        raise ValueError(f"Logger name '{name}' must start with root logger name '{ROOT_LOGGER_NAME}'")
    logger = logging.getLogger(name)
    return logger


def get_debug_logger():
    logger = logging.getLogger('onetick.py.otq_debug_mode')
    logger.setLevel("DEBUG" if config.otq_debug_mode else "INFO")
    return logger
